"""
Estimation of Historic Portfolio Volatility.
"""

import logging
import warnings

import functools
from typing import Dict, List, Optional
from typing import Callable, Tuple, Any, Union
from numbers import Number

import numpy as np
import pandas as pd
from macrosynergy.panel.historic_vol import expo_weights
from macrosynergy.management.types import NoneType, QuantamentalDataFrame
from macrosynergy.management.constants import FFILL_LIMITS, ANNUALIZATION_FACTORS
from macrosynergy.management.utils import (
    _map_to_business_day_frequency,
    get_sops,
    is_valid_iso_date,
    reduce_df,
    # standardise_dataframe,
    # ticker_df_to_qdf,
)

RETURN_SERIES_XCAT = "_PNL_USD1S_ASD"


logger = logging.getLogger(__name__)

cache = functools.lru_cache(maxsize=None)


@cache
def flat_weights_arr(lback_periods: int, *args, **kwargs) -> np.ndarray:
    """Flat weights for the look-back period."""
    return np.ones(lback_periods) / lback_periods


@cache
def expo_weights_arr(lback_periods: int, half_life: int, *args, **kwargs) -> np.ndarray:
    """Exponential weights for the lookback period."""
    return expo_weights(lback_periods=lback_periods, half_life=half_life)


def _weighted_covariance(
    x: np.ndarray,
    y: np.ndarray,
    weights_func: Callable[[int, int], np.ndarray],
    lback_periods: int,
    half_life: int,
) -> float:
    """
    Estimate covariance between two series after applying weights.

    """
    assert half_life > 0, "half_life must be greater than 0"
    assert lback_periods > 0 or lback_periods == -1, "lback_periods must be >0"
    assert x.ndim == 1 or x.shape[1] == 1, "`x` must be a 1D array or a column vector"
    assert y.ndim == 1 or y.shape[1] == 1, "`y` must be a 1D array or a column vector"
    assert x.shape[0] == y.shape[0], "`x` and `y` must have same length"

    # if either of x or y is all NaNs, return NaN
    if np.isnan(x).all() or np.isnan(y).all():
        return np.nan

    wmask = np.isnan(x) | np.isnan(y)
    weightslen = min(sum(~wmask), lback_periods if lback_periods > 0 else len(x))

    # drop NaNs and only consider the most recent lback_periods
    x, y = x[~wmask][-weightslen:], y[~wmask][-weightslen:]

    if len(x) < weightslen or weightslen == 0:
        return np.nan

    assert x.shape[0] == weightslen
    w: np.ndarray = weights_func(lback_periods=weightslen, half_life=half_life)
    x_wsum, y_wsum = (w * x).sum(), (w * y).sum()
    rss = (x - x_wsum) * (y - y_wsum)

    return w.T.dot(rss)


def estimate_variance_covariance(
    piv_ret: pd.DataFrame,
    remove_zeros: bool,
    weights_func: Callable[[int, int], np.ndarray],
    lback_periods: int,
    half_life: int,
) -> pd.DataFrame:
    """
    Estimation of the variance-covariance matrix needs to have the following
    configuration options

        1. Absolutely vs squared deviations,

        2. Flat weights (equal) vs. exponential weights,

        3. Frequency of estimation (daily, weekly, monthly, quarterly) and their weights.
    """

    cov_mat = np.zeros((len(piv_ret.columns), len(piv_ret.columns)))
    logger.info(f"Estimating variance-covariance matrix for {piv_ret.columns}")

    if remove_zeros:
        piv_ret = piv_ret.replace(0, np.nan)

    for i_b, c_b in enumerate(piv_ret.columns):
        for i_a, c_a in enumerate(piv_ret.columns[: i_b + 1]):
            logger.debug(f"Estimating covariance between {c_a} and {c_b}")
            est_vol = _weighted_covariance(
                x=piv_ret[c_a].values,
                y=piv_ret[c_b].values,
                weights_func=weights_func,
                lback_periods=lback_periods,
                half_life=half_life,
            )
            cov_mat[i_a, i_b] = cov_mat[i_b, i_a] = est_vol

    assert np.all((cov_mat.T == cov_mat) ^ np.isnan(cov_mat))

    return pd.DataFrame(cov_mat, index=piv_ret.columns, columns=piv_ret.columns)


def _downsample_returns(
    piv_df: pd.DataFrame,
    freq: str = "m",
) -> pd.DataFrame:
    # TODO create as a general convert_frequency function
    # TODO current aggregator is `art` (check definition of name in R code)
    # TODO test [1] input data is daily and [2] daily gives daily output

    freq = _map_to_business_day_frequency(freq)
    # TODO we should fix why we get the warnings...
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        piv_new_freq: pd.DataFrame = (
            (1 + piv_df / 100).resample(freq).prod() - 1
        ) * 100
        warnings.resetwarnings()
    return piv_new_freq


def get_max_lookback(lb: int, nt: float) -> int:
    """
    Calculate the maximum lookback period for a given lookback period and nan tolerance.

    Parameters
    ----------
    lb : int
        the lookback period.
    nt : float
        the nan tolerance.

    Returns
    -------
    int
        the maximum lookback period.
    """

    return int(np.ceil(lb * (1 + nt))) if lb > 0 else 0


def _calculate_multi_frequency_vcv_for_period(
    pivot_returns: pd.DataFrame,
    pivot_signals: pd.DataFrame,
    rebal_date: pd.Timestamp,
    est_freqs: List[str],
    est_weights: List[float],
    weights_func: Callable[[int, int], np.ndarray],
    lback_periods: List[int],
    half_life: List[int],
    nan_tolerance: float,
    remove_zeros: bool,
) -> pd.DataFrame:

    window_df = pivot_returns.loc[pivot_returns.index <= rebal_date]
    dict_vcv: Dict[str, pd.DataFrame] = {}

    for freq, lb, hl in zip(est_freqs, lback_periods, half_life):
        piv_ret = _downsample_returns(window_df, freq=freq).iloc[
            -get_max_lookback(lb=lb, nt=nan_tolerance) :
        ]
        dict_vcv[freq] = estimate_variance_covariance(
            piv_ret=piv_ret,
            lback_periods=lb,
            remove_zeros=remove_zeros,
            weights_func=weights_func,
            half_life=hl,
        )
        # if dict_vcv[freq].isna().any().any():
        #     raise ValueError(
        #         f"N/A values in variance-covariance matrix at freq={freq} at real_date={rebal_date}!\n"
        #         f"{dict_vcv[freq].isna().any()}"
        #     )

    # NOTE: in this case Float+NA = Na
    vcv_df: pd.DataFrame = sum(
        [
            est_weights[ix] * ANNUALIZATION_FACTORS[freq] * dict_vcv[freq]
            for ix, freq in enumerate(est_freqs)
        ]
    )

    return vcv_df


def _calc_vol_tuple(
    vcv_df: pd.DataFrame,
    signals: pd.DataFrame,
    date: pd.Timestamp,
    available_cids: List[str],
) -> Tuple[pd.Timestamp, float]:

    s = signals.loc[date, :].copy()

    s = s.loc[available_cids]
    vcv_df = vcv_df.loc[available_cids, available_cids]
    if not set(s.index) == set(vcv_df.columns):
        raise ValueError(
            "Signals and variance-covariance matrix do not have the same columns."
            f"\nSignals: {s.columns.tolist()}"
            f"\nVariance-Covariance: {vcv_df.columns.tolist()}"
        )

    idx_mask = s.isna() | (s.abs() < 1e-6)
    s.loc[idx_mask] = 0
    vcv_df.loc[idx_mask, :] = 0
    vcv_df.loc[:, idx_mask] = 0
    assert not vcv_df.isna().any().any(), "N/A values in variance-covariance matrix!\n"

    pvol: float = np.sqrt(s.T.dot(vcv_df).dot(s))
    return date, pvol


def stack_covariances(
    vcv_df: pd.DataFrame,
    real_date: pd.Timestamp,
) -> pd.DataFrame:
    """Stack the covariance matrix DataFrame."""
    return (
        vcv_df.rename_axis("fid1", axis=0)
        .rename_axis("fid2", axis=1)
        .stack()
        .to_frame("value")
        .reset_index()
        .assign(real_date=real_date)
    )


def _get_first_usable_date(
    pivot_returns: pd.DataFrame,
    pivot_signals: pd.DataFrame,
    rebal_dates: pd.Series,
    est_freqs: List[str],
    lback_periods: List[int],
    nan_tolerance: float,
) -> pd.Series:
    max_lb = 0
    # for each frequency and lookback
    for lb, est_freq in zip(lback_periods, est_freqs):
        _max_lb = get_max_lookback(lb, nan_tolerance)
        _max_lb = (
            FFILL_LIMITS[_map_to_business_day_frequency(est_freq)]
            if _max_lb == 0
            else _max_lb
        )
        max_lb = _max_lb if _max_lb > max_lb else max_lb

    assert set(pivot_returns.columns.tolist()) == set(pivot_signals.columns.tolist())
    pr_starts = {}
    ps_starts = {}
    for col in pivot_returns.columns.tolist():
        # 'full' start date for returns - where the maximum lookback period is available
        fstart_ret = pivot_returns[col].first_valid_index() + pd.offsets.BDay(max_lb)
        fstart_sig = pivot_signals[col].first_valid_index() + pd.offsets.BDay(max_lb)
        pr_starts[col] = rebal_dates[rebal_dates >= fstart_ret].min()
        ps_starts[col] = rebal_dates[rebal_dates >= fstart_sig].min()

    # get the later of the two start dates and return
    return pd.Series(
        {k: max(pr_starts[k], ps_starts[k]) for k in pr_starts.keys()},
        name="real_date",
    )


def _calculate_portfolio_volatility(
    pivot_returns: pd.DataFrame,
    pivot_signals: pd.DataFrame,
    rebal_freq: str,
    est_freqs: List[str],
    est_weights: List[float],
    weights_func: Callable[[int, int], np.ndarray],
    lback_periods: List[int],
    half_life: List[int],
    nan_tolerance: float,
    remove_zeros: bool,
    portfolio_return_name: str,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    logger.info(
        f"Calculating portfolio volatility "
        f"for FIDS={pivot_returns.columns.tolist()} "
        f"from {min(pivot_returns.index.min(), pivot_signals.index.min())} "
        f"to {max(pivot_returns.index.max(), pivot_signals.index.max())}, with "
        f"lback_periods={lback_periods}, nan_tolerance={nan_tolerance}, "
        f"remove_zeros={remove_zeros}, rebal_freq={rebal_freq}, est_freqs={est_freqs}, "
        f"est_weights={est_weights} "
    )

    rebal_dates = get_sops(dates=pivot_signals.index, freq=rebal_freq)

    # Returns batches
    logger.info(
        "Rebalance portfolio from %s to %s (%s times)",
        rebal_dates.min(),
        rebal_dates.max(),
        rebal_dates.shape[0],
    )

    # td = rebal_dates.iloc[-1]

    # TODO convert frequencies
    list_vcv: List[pd.DataFrame] = []
    list_pvol: List[Tuple[pd.Timestamp, np.float64]] = []
    first_starts = _get_first_usable_date(
        pivot_returns=pivot_returns,
        pivot_signals=pivot_signals,
        rebal_dates=rebal_dates,
        est_freqs=est_freqs,
        lback_periods=lback_periods,
        nan_tolerance=nan_tolerance,
    )

    for td in rebal_dates:
        avails = first_starts[first_starts <= td].index.tolist()
        if len(avails) == 0:
            logger.warning(
                f"No data available for {td} with lookback period of {max(lback_periods)} days."
            )
            continue
        vcv_df = _calculate_multi_frequency_vcv_for_period(
            pivot_returns=pivot_returns[avails],
            pivot_signals=pivot_signals[avails],
            rebal_date=td,
            est_freqs=est_freqs,
            est_weights=est_weights,
            weights_func=weights_func,
            lback_periods=lback_periods,
            half_life=half_life,
            nan_tolerance=nan_tolerance,
            remove_zeros=remove_zeros,
        )

        list_vcv.append(stack_covariances(vcv_df=vcv_df, real_date=td))
        vol_tuple = _calc_vol_tuple(
            vcv_df=vcv_df,
            # signals=signals,
            signals=pivot_signals,
            date=td,
            available_cids=avails,
        )
        list_pvol.append(vol_tuple)

    pvol = pd.DataFrame(
        list_pvol,
        columns=["real_date", portfolio_return_name],
    ).set_index("real_date")

    vcv_df_long = pd.concat(list_vcv, axis=0)  # add to cls.vcv

    vcv_df_long["helper"] = vcv_df_long[["fid1", "fid2", "real_date"]].apply(
        func=(lambda x: "-".join(sorted([x["fid1"], x["fid2"]])) + str(x["real_date"])),
        axis=1,
    )
    vcv_df_long = (
        vcv_df_long.drop_duplicates(subset=["helper"])
        .drop(columns=["helper"])
        .reset_index(drop=True)
    )

    return pvol, vcv_df_long


def _hist_vol(
    pivot_signals: pd.DataFrame,
    pivot_returns: pd.DataFrame,
    sname: str,
    rebal_freq: str,
    lback_meth: str,  # TODO allow for different method at different frequencies
    lback_periods: List[int],  # default all for all
    half_life,
    est_freqs: List[str],
    est_weights: List[float],
    nan_tolerance: float,
    remove_zeros: bool,
    return_variance_covariance: bool,
) -> List[pd.DataFrame]:
    """
    Calculates historic volatility for a given strategy. It assumes that the dataframe
    is composed solely of the relevant signals and returns for the strategy.

    Parameters
    ----------
    pivot_signals : pd.DataFrame
        the pivot table of the contract signals.
    pivot_returns : pd.DataFrame
        the pivot table of the contract returns.
    rebal_freq : str
        the frequency of the volatility estimation. Default is 'm' for monthly.
        Alternatives are 'w' for business weekly, 'd' for daily, and 'q' for quarterly.
        Estimations are conducted for the end of the period.
    lback_periods : int
        the number of periods to use for the lookback period of the volatility-targeting
        method. Default is 21.
    lback_meth : str
        the method to use for the lookback period of the volatility-targeting method.
        Default is 'ma' for moving average. Alternative is "xma", for exponential moving
        average.
    half_life : int
        Refers to the half-time for "xma" and full lookback period for "ma". Default is
        11.
    nan_tolerance : float
        maximum ratio of NaNs to non-NaNs in a lookback window, if exceeded the
        resulting volatility is set to NaN. Default is 0.25.
    remove_zeros : bool
        removes zeroes as invalid entries and shortens the effective window.
    """

    lback_meth = lback_meth.lower()
    if lback_meth not in ["ma", "xma"]:
        raise NotImplementedError(
            f"`lback_meth` must be 'ma' or 'xma'; got {lback_meth}"
        )

    # TODO get the correct rebalance dates
    weights_func = flat_weights_arr if lback_meth == "ma" else expo_weights_arr
    logger.info(
        "Found lback_meth=%s, using weights_func=%s", lback_meth, weights_func.__name__
    )
    portfolio_return_name = f"{sname}{RETURN_SERIES_XCAT}"

    pvol_df: pd.DataFrame
    vcv_df: pd.DataFrame
    pvol_df, vcv_df = _calculate_portfolio_volatility(
        pivot_returns=pivot_returns,
        pivot_signals=pivot_signals,
        rebal_freq=rebal_freq,
        weights_func=weights_func,
        portfolio_return_name=portfolio_return_name,
        lback_periods=lback_periods,
        remove_zeros=remove_zeros,
        nan_tolerance=nan_tolerance,
        half_life=half_life,
        est_freqs=est_freqs,
        est_weights=est_weights,
    )

    # assert portfolio_return_name the only column
    pvol_df = pvol_df.reset_index()
    assert set(pvol_df.columns.tolist()) == set([portfolio_return_name, "real_date"])

    nan_dates = pvol_df[pvol_df[portfolio_return_name].isna()]["real_date"].copy()
    if len(nan_dates) > 0:
        logger.warning(
            f"Found NaNs in {portfolio_return_name} at: {nan_dates.tolist()}, dropping all NaNs."
        )
        pvol_df = pvol_df[~pvol_df["real_date"].isin(nan_dates)].copy()

    pvol_df = pvol_df.set_index("real_date")

    if return_variance_covariance:
        return [pvol_df, vcv_df]
    return [pvol_df]


def unstack_covariances(
    vcv_df: pd.DataFrame,
    fillna: bool = False,
) -> Dict[str, pd.DataFrame]:
    """Unstack the covariance matrix DataFrame."""
    vcvs: Dict[str, pd.DataFrame] = {}
    for dt, df in vcv_df.groupby("real_date"):
        vcv = df.pivot(index="fid2", columns="fid1", values="value")
        if fillna:
            vcv = vcv.fillna(vcv.T)
            assert all(vcv == vcv.T)
        vcvs[pd.Timestamp(dt).strftime("%Y-%m-%d")] = vcv

    return vcvs


def _check_input_arguments(
    arguments: List[Tuple[Any, str, Union[type, Tuple[type, type]]]]
):
    # TODO move to general utils
    for varx, namex, typex in arguments:
        if not isinstance(varx, typex):
            raise TypeError(f"`{namex}` must be {typex}.")
        if typex in [str, list, dict] and len(varx) == 0:
            raise ValueError(f"`{namex}` must not be an empty {str(typex)}.")


def _check_frequency(freq: str, freq_type: str):
    # TODO move to general utils
    try:
        _map_to_business_day_frequency(freq)
    except ValueError as e:
        raise ValueError(
            f"`{freq_type:s}` ({freq:s}) must be a valid frequency string: {e}"
        )


def _check_missing_data(
    df: pd.DataFrame, sname: str, fids: List[str], rstring: str
) -> None:
    ## Check that there is atleast one contract signal for the strategy
    if not any(df["ticker"].str.endswith(f"_CSIG_{sname}")):
        raise ValueError(f"No contract signals for strategy `{sname}`.")

    u_tickers: List[str] = list(df["ticker"].unique())
    for contx in fids:
        if not any(
            [tx.startswith(contx) and tx.endswith(f"_CSIG_{sname}") for tx in u_tickers]
        ):
            raise ValueError(f"Contract identifier `{contx}` not in dataframe.")

    if not all([f"{contx}{rstring}" in u_tickers for contx in fids]):
        missing_tickers = [
            f"{contx}{rstring}"
            for contx in fids
            if f"{contx}{rstring}" not in u_tickers
        ]
        raise ValueError(
            f"The dataframe is missing the following return series: {missing_tickers}"
        )


def _check_est_args(
    est_freqs: List[str],
    est_weights: List[Number],
    lback_periods: List[int],
    half_life: List[int],
) -> Tuple[List[str], List[float], List[int], List[int]]:
    # Calculate the maximum length of the provided lists
    max_len = max(len(est_freqs), len(est_weights), len(lback_periods), len(half_life))

    def expand_list(lst, name):
        if len(lst) == 1:
            return lst * max_len
        elif len(lst) != max_len:
            raise ValueError(
                "All lists must have length 1 or the same length as the longest "
                f"list ({max_len}). '{name}' has length {len(lst)}."
            )
        return lst

    # Expand lists to match the maximum length
    est_freqs = expand_list(est_freqs, "est_freqs")
    est_weights = expand_list(est_weights, "est_weights")
    lback_periods = expand_list(lback_periods, "lback_periods")
    half_life = expand_list(half_life, "half_life")

    inv_weights_msg = "Invalid weights in `est_weights` at index {ix:d}"
    inv_lback_msg = "Invalid lookback period in `lback_periods` at index {ix:d}: {lb:d}"
    inv_hl_msg = "Invalid half-life in `half_life` at index {ix:d}: {hl:d}"

    for ix, (freq, weight, lback, hl) in enumerate(
        zip(est_freqs, est_weights, lback_periods, half_life)
    ):
        _check_frequency(freq=freq, freq_type=f"est_freq[{ix:d}]")

        if not isinstance(weight, Number) or weight < 0:
            raise ValueError(inv_weights_msg.format(ix=ix))

        # stated idiosyncratically to allow for -1
        if not isinstance(lback, int) or (lback < 0 and lback != -1):
            raise ValueError(inv_lback_msg.format(ix=ix, lb=lback))

        if not isinstance(hl, int) or hl < 0:
            raise ValueError(inv_hl_msg.format(ix=ix, hl=hl))

    # normalize est_weights
    if not np.isclose(np.sum(est_weights), 1):
        est_weights = list(np.array(est_weights) / np.sum(est_weights))

    return est_freqs, est_weights, lback_periods, half_life


def add_fid_column(df: QuantamentalDataFrame, rstring: str) -> QuantamentalDataFrame:
    """Add financial identifier (fid) to DataFrame."""
    df["fid"] = (
        df["cid"].astype(str)
        + "_"
        + df["xcat"]
        .str.split("_")
        .map(
            lambda x: (
                x[0][: -len(rstring.split("_")[0])]
                if x[0].endswith(rstring.split("_")[0])
                else x[0]
            )
        )
    )
    return df


def historic_portfolio_vol(
    df: pd.DataFrame,
    sname: str,
    fids: List[str],
    rstring: str = "XR",
    rebal_freq: str = "m",
    lback_meth: str = "ma",
    est_freqs: Union[str, List[str]] = ["D", "W", "M"],  # "m", "w", "d", "q"
    est_weights: Union[Number, List[Number]] = [1, 1, 1],  # default equal weights
    lback_periods: Union[int, List[int]] = [-1, -1, -1],  # default all for all
    half_life: Union[int, List[int]] = [11, 5, 6],
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Optional[dict] = None,
    nan_tolerance: float = 0.25,
    remove_zeros: bool = True,
    return_variance_covariance: bool = True,
) -> Union[QuantamentalDataFrame, Tuple[QuantamentalDataFrame, pd.DataFrame]]:
    """
    Historical portfolio volatility.  Estimates annualized standard deviations of a
    portfolio, based on historic variances and co-variances.

    Parameters
    ----------
    df : QuantamentalDataFrame
        JPMaQS standard DataFrame containing contract-specific signals and return
        series.
    sname : str
        the name of the strategy. It must correspond to contract signals in the
        dataframe, which have the format "<cid>_<ctype>_CSIG_<sname>", and which are
        typically calculated by the function contract_signals().
    fids : List[str]
        list of financial contract identifiers in the format "<cid>_<ctype>". It must
        correspond to contract signals in the dataframe.
    rstring : str
        a general string of the return category. This identifies the contract returns
        that are required for the volatility-targeting method, based on the category
        identifier format <cid>_<ctype><rstring> in accordance with JPMaQS conventions.
        Default is 'XR'.
    rebal_freq : str
        the frequency of rebalancing and volatility estimation. Default is 'M' for
        monthly. Alternatives are 'W' for business weekly, 'D' for daily, and 'Q' for
        quarterly. Estimations are conducted for the end of the period.
    est_freqs : List[str]
        the list of frequencies for which the volatility is estimated. Volatility for a
        given period is the weighted sum of the volatilities estimated for each frequency.
        Default is ["D", "W", "M"].
    est_weights : List[float]
        the list of weights for each frequency in `est_freqs`. Weights are normalized
        before applying. In cases where there may be missing data or NaNs in the result, the
        remaining weights are normalized. Default is None, which means that the weights are
        equal.
    lback_meth : str
        the method to use for the lookback period of the volatility-targeting method.
        Default is "ma" for moving average. Alternative is "xma", for exponential moving
        average.
    lback_periods : List[int]
        the number of periods to use for the lookback period of the volatility-targeting
        method. Each element corresponds to the the same index in `est_freqs`. Passing a
        single element will apply the same value to all frequencies. Default is [-1], which
        means that the lookback period is the full available data for all specified
        frequencies.
    half_life : List[int]
        number of periods in the half-life of the exponential moving average. Each
        element corresponds to the same index in `est_freqs`.
    start : str
        the start date of the data. Default is None, which means that the start date is
        taken from the dataframe.
    end : str
        the end date of the data. Default is None, which means that the end date is
        taken from the dataframe.
    blacklist : dict
        a dictionary of contract identifiers to exclude from the calculation. Default is
        None, which means that no contracts are excluded.
    nan_tolerance : float
        maximum ratio of number of NaN values to the total number of values in a
        lookback window. If exceeded the resulting volatility is set to NaN, else prior non-
        zero values are added to the window instead. Default is 0.25.
    remove_zeros : bool
        if True (default) any returns that are exact zeros will not be included in the
        lookback window and prior non-zero values are added to the window instead.

    Returns
    -------
    pd.DataFrame
        JPMaQS dataframe of annualized standard deviation of estimated strategy PnL,
        with category name <sname>_PNL_USD1S_ASD. TODO: check if this is correct. The values
        are in % annualized. Values between estimation points are forward filled.

    Notes
    -----
    If returns in the lookback window are not available the function will replace them with
    the average of the available returns of the same contract type. If no returns are
    available for a contract type the function will reduce the lookback window up to a
    minimum of 11 days. If no returns are available for a contract type for at least 11
    days the function returns an NaN for that date and sends a warning of all the dates
    for which this happened.
    """

    if isinstance(lback_periods, Number):
        lback_periods = [lback_periods]
    if isinstance(half_life, Number):
        half_life = [half_life]
    if isinstance(est_weights, Number):
        est_weights = [est_weights]
    if isinstance(est_freqs, str):
        est_freqs = [est_freqs]

    ## Check inputs
    # TODO create function for this? Also, do we want to create the set of failures (not just first one)?
    _check_input_arguments(
        arguments=[
            (sname, "sname", str),
            (fids, "fids", list),
            (rstring, "rstring", str),
            (rebal_freq, "rebal_freq", str),
            (lback_meth, "lback_meth", str),
            (lback_periods, "lback_periods", list),
            (half_life, "half_life", list),
            (est_freqs, "est_freqs", list),
            (est_weights, "est_weights", list),
            (start, "start", (str, NoneType)),
            (end, "end", (str, NoneType)),
            (blacklist, "blacklist", (dict, NoneType)),
            (nan_tolerance, "nan_tolerance", float),
            (remove_zeros, "remove_zeros", bool),
            (return_variance_covariance, "return_variance_covariance", bool),
        ]
    )

    # Check the frequency arguments
    _check_frequency(freq=rebal_freq, freq_type="rebal_freq")

    for ix, freq in enumerate(est_freqs):
        _check_frequency(freq=freq, freq_type=f"est_freq[{ix:d}]")

    ## Check estimation frequency weights
    est_freqs, est_weights, lback_periods, half_life = _check_est_args(
        est_freqs=est_freqs,
        est_weights=est_weights,
        lback_periods=lback_periods,
        half_life=half_life,
    )

    ## Standardize and copy DF
    df = QuantamentalDataFrame(df)
    rebal_freq = _map_to_business_day_frequency(rebal_freq)
    est_freqs: List[str] = [_map_to_business_day_frequency(freq) for freq in est_freqs]

    ## Check the dates
    if start is None:
        start: str = pd.Timestamp(df["real_date"].min()).strftime("%Y-%m-%d")

    if end is None:
        end: str = pd.Timestamp(df["real_date"].max()).strftime("%Y-%m-%d")

    for dx, nx in [(start, "start"), (end, "end")]:
        if not is_valid_iso_date(dx):
            raise ValueError(f"`{nx}` must be a valid ISO-8601 date string")

    ## Reduce the dataframe
    df: pd.DataFrame = reduce_df(df=df, start=start, end=end, blacklist=blacklist)
    df = QuantamentalDataFrame(df).add_ticker_column()
    u_tickers: List[str] = df.list_tickers()

    ## Check for missing data
    _check_missing_data(df=df, sname=sname, fids=fids, rstring=rstring)

    # Add financial identifier (fid) to DataFrame
    df = add_fid_column(df=df, rstring=rstring)

    ## Filter out data-frame and select contract signals (CSIG) and returns (XR)
    filt_csigs: List[str] = [tx for tx in u_tickers if tx.endswith(f"_CSIG_{sname}")]
    filt_xrs: List[str] = [tx for tx in u_tickers if tx.endswith(rstring)]

    # TODO check if all exists

    pivot_signals: pd.DataFrame = df.loc[df["ticker"].isin(filt_csigs)].pivot(
        index="real_date", columns="fid", values="value"
    )

    pivot_returns: pd.DataFrame = df.loc[df["ticker"].isin(filt_xrs)].pivot(
        index="real_date", columns="fid", values="value"
    )
    assert set(pivot_signals.columns) == set(pivot_returns.columns)

    result: List[pd.DataFrame] = _hist_vol(
        pivot_returns=pivot_returns,
        pivot_signals=pivot_signals,
        sname=sname,
        rebal_freq=rebal_freq,
        est_freqs=est_freqs,
        est_weights=est_weights,
        lback_periods=lback_periods,
        lback_meth=lback_meth,
        half_life=half_life,
        nan_tolerance=nan_tolerance,
        remove_zeros=remove_zeros,
        return_variance_covariance=return_variance_covariance,
    )

    assert len(result) == 1 + int(return_variance_covariance)

    result[0] = QuantamentalDataFrame.from_wide(df=result[0])
    if return_variance_covariance:
        return result[0], result[1]
    return result[0]


if __name__ == "__main__":
    from macrosynergy.management.simulate import simulate_returns_and_signals

    np.random.seed(42)  # Fix numpy seed to 42 for reproducibility

    # Signals: FXCRY_NSA, EQCRY_NSA (rename to FX_CSIG_STRAT, EQ_CSIG_STRAT)
    # Returns: FXXR_NSA, EQXR_NSA (renamed to FXXR, EQXR)
    cids: List[str] = ["EUR", "GBP", "AUD", "CAD"]
    xcats: List[str] = ["EQ"]
    ctypes = xcats.copy()
    start: str = "2000-01-01"
    xr_tickers = [f"{cid}_{xcat}XR" for cid in cids for xcat in xcats]
    cs_tickers = [f"{cid}_{xcat}_CSIG_STRAT" for cid in cids for xcat in xcats]
    fids: List[str] = [f"{cid}_{ctype}" for cid in cids for ctype in ctypes]

    df = simulate_returns_and_signals(
        cids=cids,
        xcat=xcats[0],
        return_suffix="XR",
        signal_suffix="CSIG_STRAT",
        start=start,
        years=20,
    )
    # TODO simulate_returns_and_signals are risk-signals, not contract signals. We need to adjust for volatility and common (observed) factor.
    end = df["real_date"].max().strftime("%Y-%m-%d")

    df_copy = df.copy()  # TODO why copy?

    N_p_nans = 0.01
    df["value"] = df["value"].apply(
        lambda x: x if np.random.rand() > N_p_nans else np.nan
    )

    df_vol, vcv_df = historic_portfolio_vol(
        df=df,
        sname="STRAT",
        fids=fids,
        rebal_freq="m",
        est_freqs=["D", "W", "M"],
        est_weights=[0.1, 0.2, 0.7],
        lback_periods=[30, 20, -1],
        half_life=[10, 5, 2],
        lback_meth="xma",
        rstring="XR",
        start=start,
        end=end,
        return_variance_covariance=True,
    )

    vcvs_dict = unstack_covariances(vcv_df)
    dates = [
        dt.strftime("%Y-%m-%d")
        for dt in sorted(pd.to_datetime(list(vcvs_dict.keys())))[-9:]
    ]
    # with sns.axes_style("whitegrid"):
    #     fig, ax = plt.subplots(3, 3, figsize=(15, 15))
    #     for ix, dt in enumerate(dates):
    #         sns.heatmap(vcvs_dict[dt], ax=ax[ix // 3, ix % 3])
    #         ax[ix // 3, ix % 3].set_title(dt)
    #     plt.tight_layout()
    #     plt.show()

    df_copy_vol: pd.DataFrame = historic_portfolio_vol(
        df=df_copy,
        sname="STRAT",
        fids=fids,
        rebal_freq="m",
        lback_periods=15,
        lback_meth="ma",
        half_life=11,
        rstring="XR",
        start=start,
        end=end,
        return_variance_covariance=False,
    )

    # print(df_copy_vol.head(10))
    # print(df_copy_vol.tail(10))



'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.2.0

version: 1.2.0

full_version: 1.2.0

git_revision: b3f53f5bc7db343d3083771e213e950374f6ad00

release: True

'''
short_version: str = '1.2.0'
version: str = '1.2.0'
full_version: str = '1.2.0'
git_revision: str = 'b3f53f5bc7db343d3083771e213e950374f6ad00'
release: bool = True
if not release:
    version = full_version



from .panel_selectors import (
    BasePanelSelector,
    LarsSelector,
    LassoSelector,
    MapSelector,
)

from .scalers import (
    BasePanelScaler,
    PanelMinMaxScaler,
    PanelStandardScaler,
)

from .transformers import (
    ZnScoreAverager,
    PanelPCA,
)

__all__ = [
    # selectors
    "BasePanelSelector",
    "LarsSelector",
    "LassoSelector",
    "MapSelector",
    # scalers
    "BasePanelScaler",
    "PanelMinMaxScaler",
    "PanelStandardScaler",
    # transformers
    "PanelPCA",
    "ZnScoreAverager",
]


"""
Module with functionality for generating mock quantamental data vintages for testing
purposes.
"""

import numpy as np
import pandas as pd
import math
import datetime as dt
from typing import List
from macrosynergy.management.utils import _map_to_business_day_frequency
from macrosynergy.management.constants import ANNUALIZATION_FACTORS


class VintageData:
    """
    Creates standardized dataframe of single-ticker vintages. This class creates
    standardized grade 1 and grade 2 vintage data.

    Parameters
    ----------
    ticker : str
        ticker name
    cutoff : str
        last possible release date. The format must be '%Y-%m-%d'. All other dates are
        calculated from this one. Default is end 2020.
    release_lags : list
        list of integers in ascending order denoting lags of the first, second etc.
        release in (calendar) days. Default is first release after 15 days and revision
        after 30 days. If days fall on weekend they will be delayed to Monday.
    number_firsts : int
        number of first-release vintages in the simulated data set. Default is 24.
    shortest : int
        number of observations in the first (shortest) vintage. Default is 36.
    freq : str
        letter denoting the frequency of the vintage data. Must be one of 'M' (monthly,
        default), 'Q' (quarterly)  or 'W' (weekly).
    start_value : float
        expected first value of the random series. Default is 100.
    trend_ar : float
        annualized trend. Default is 5% linear drift per year. This is applied to the
        start value. If the start value is not positive the linear trend is added as number.
    sd_ar : float
        annualized standard deviation. Default is sqrt(12).
    seasonal : float
        adds seasonal pattern (applying linear factor from low to high through the year)
        with value denoting the average % seasonal factor through the year. Default is None.
        The seasonal pattern makes only sense for values that are strictly positive and are
        interpreted as indices.
    added_dates : int
        number of added first release dates, used for grade 2 dataframe generation.
        Default is 12.
    """

    def __init__(
        self,
        ticker,
        cutoff="2020-12-31",
        release_lags=[15, 30],
        number_firsts=24,
        shortest=36,
        freq="M",
        start_value=100,
        trend_ar=5,
        sd_ar=math.sqrt(12),
        seasonal=None,
        added_dates=12,
    ):
        self.ticker = ticker

        self.cutoff = self.date_check(cutoff)
        self.release_lags = release_lags
        self.number_firsts = number_firsts
        self.shortest = shortest

        self.freq = _map_to_business_day_frequency(freq)
        self.freq_int = ANNUALIZATION_FACTORS
        self.af = self.freq_int[freq]

        self.start_value = start_value
        self.trend_ar = trend_ar
        self.sd_ar = sd_ar
        self.seasonal = seasonal

        self.number_firsts_gr2 = self.number_firsts + added_dates
        self.dates = set()

    @staticmethod
    def date_check(date_string):
        """
        Validates that the dates passed are valid timestamp expressions and will convert
        to the required form '%Y-%m-%d'.

        Parameters
        ----------
        date_string : str
            valid date expression. For instance, "1st January, 2000."

        Raises
        ------
        TypeError
            if the date_string is not a string.
        ValueError
            if the date_string is not in the correct format.
        """

        date_error = "Expected form of string: '%Y-%m-%d'."
        if date_string is not None:
            if not isinstance(date_string, str):
                raise TypeError("`date_string` must be a string.")
            try:
                pd.Timestamp(date_string).strftime("%Y-%m-%d")
            except ValueError:
                raise ValueError(date_error)
            else:
                date = pd.Timestamp(date_string).strftime("%Y-%m-%d")
                return pd.Timestamp(date)

    @staticmethod
    def week_day(rel_date, day):
        if day == 0:
            return rel_date - dt.timedelta(days=1)

        return rel_date + dt.timedelta(days=1)

    def seasonal_adj(self, obs_dates, seas_factors, values):
        """
        Method used to seasonally adjust the series. Economic data can vary according to
        the season.

        Parameters
        ----------
        obs_dates : List[pd.Timestamps]
            observation dates for the series.
        seas_factors : List[float]
            seasonal factors.
        values : List[float]
            existing values that have not been seasonally adjusted.

        Returns
        -------
        List[float]
            returns a list of values which have been adjusted seasonally
        """

        if self.freq == "W":
            week_dates = obs_dates.isocalendar().week
            condition = np.where(week_dates > 52)[0]
            if condition.size > 0:
                index = next(iter(condition))
                week_dates[index] = 52

            week = (week_dates - 1).to_numpy().astype(dtype=np.uint8)
            seasonal = seas_factors[week]
            return values * (1 + (seasonal / 100))

        month = 12 / self.af
        month_freq = obs_dates.month // month
        month_freq = month_freq.astype(dtype=np.uint8)

        return values * (1 + seas_factors[month_freq - 1] / 100)

    def make_grade1(self):
        ref_date = self.cutoff - dt.timedelta(self.release_lags[0])
        ref_date = ref_date.replace(day=1)

        eop_dates = pd.date_range(
            end=ref_date, periods=self.number_firsts, freq=self.freq
        )
        eop_dates = eop_dates.date
        vin_lengths = list(range(self.shortest, self.shortest + self.number_firsts))
        v_first = vin_lengths[0]

        df_gr1 = pd.DataFrame(columns=["release_date", "observation_date", "value"])
        obs_dates = pd.date_range(end=eop_dates[0], periods=v_first, freq=self.freq)

        list_ = np.linspace(0, (self.af - 1), self.af)
        linear_scale = list_ - np.mean(list_)

        seas_factors = self.seasonal / np.std(linear_scale)
        seas_factors *= linear_scale
        df_rels: List[pd.DataFrame] = []
        for i, eop_date in enumerate(eop_dates):
            v = vin_lengths[i]
            if i > 0:
                length = vin_lengths[i - 1]
                date = pd.Timestamp(eop_date)
                obs_dates = obs_dates.insert(loc=length, item=date)

            data = np.linspace(0, (v - 1), v)
            if self.start_value > 0:
                data = (data * (self.trend_ar / self.af)) / 100
                data = 1 + data
                trend = self.start_value * data
            else:
                data = data * (self.trend_ar / self.af)
                trend = self.start_value + data

            for rl in self.release_lags:
                rel_date = eop_date + dt.timedelta(days=rl)
                day = rel_date.weekday() + 5
                if day >= 0:
                    rel_date = self.week_day(rel_date, day)

                self.dates.add(rel_date)
                values = trend + np.random.normal(0, self.sd_ar / math.sqrt(self.af), v)
                if self.seasonal is not None:
                    values = self.seasonal_adj(obs_dates, seas_factors, values)

                df_rel = pd.DataFrame(
                    {
                        "release_date": rel_date,
                        "observation_date": obs_dates.date,
                        "value": values,
                    }
                )
                df_rels.append(df_rel)

        df_gr1 = pd.concat(df_rels, ignore_index=True)

        df_gr1["grading"] = 1
        return self.add_ticker_parts(df_gr1)

    def make_graded(self, grading, upgrades=[]):
        """
        Simulates an explicitly graded dataframe with a column 'grading'.

        Parameters
        ----------
        grading : list
            optional addition of grading column. List of grades used from lowest to
            highest. Default is None. Must be a subset of [3, 2.2, 2.1, 1].
        upgrades : list
            indices of release dates at which the series upgrade. Must have length of
            grading minus one. Default is None.
        """

        assert len(upgrades) == (len(grading) - 1)
        df_grd = self.make_grade1()
        df_grd["grading"] = str(grading[0])
        dates = sorted(self.dates)
        for i, up in enumerate(upgrades):
            filt = df_grd["release_date"] >= dates[up]
            df_grd.loc[filt, "grading"] = str(grading[(i + 1)])

        return df_grd

    @staticmethod
    def map_weekday(date):
        day = date.weekday() - 5
        if day >= 0:
            date = VintageData.week_day(date, day)

        return date

    def make_grade2(self):
        """
        Method used to construct a dataframe that consists of each respective
        observation date and the corresponding release date(s) (the release dates are
        computed using the observation date and the time-period(s) specified in the
        field "release_lags").

        Returns
        -------
        pd.DataFrame
            Will return the DataFrame with the additional columns.
        """

        ref_date = self.cutoff - dt.timedelta(self.release_lags[0])
        ref_date = ref_date.replace(day=1)

        eop_dates = pd.date_range(
            end=ref_date, periods=self.number_firsts_gr2, freq=self.freq
        )
        eop_list = eop_dates.date
        eop_arr = np.array(eop_list)

        df_gr2 = pd.DataFrame(columns=["release_date", "observation_date"])

        n_ends = len(eop_list)
        n_release = len(self.release_lags)

        data = np.zeros((n_ends, n_release), dtype=object)

        for i in range(n_release):
            dates = eop_arr + dt.timedelta(days=self.release_lags[i])
            dates = list(map(self.map_weekday, dates))
            data[:, i] = np.array(dates)

        shape = data.shape
        data = data.reshape((shape[0] * shape[1]))
        eop_arr = np.sort(np.tile(eop_arr, n_release))

        df_gr2["release_date"] = data
        df_gr2["observation_date"] = eop_arr
        df_gr2["ticker"] = self.ticker

        return self.add_ticker_parts(df_gr2)

    def add_ticker_parts(self, df):
        """
        Method used to add the associated tickers.

        Parameters
        ----------
        df : pd.DataFrame
            standardised dataframe.

        Returns
        -------
        pd.DataFrame
            Will return the DataFrame with the additional columns.
        """

        old_cols = list(df.columns)
        add_cols = ["cross_section", "category_code", "adjustment", "transformation"]
        ticker_parts = self.ticker.split("_", 3)

        for i, tick in enumerate(ticker_parts):
            df[add_cols[i]] = tick

        if "transformation" not in df.columns:
            df["transformation"] = None

        return df.loc[:, add_cols + old_cols]


if __name__ == "__main__":
    vins_m = VintageData(
        "USD_INDX_SA",
        cutoff="2019-06-30",
        release_lags=[3, 20, 25],
        number_firsts=12,
        shortest=12,
        sd_ar=5,
        trend_ar=20,
        seasonal=10,
        added_dates=6,
    )

    dfm1 = vins_m.make_grade1()

    dfm1.groupby("release_date").agg(["mean", "count"])
    dfm2 = vins_m.make_grade2()
    dfmg = vins_m.make_graded(grading=[3, 2.1, 1], upgrades=[12, 24])

    vins_q = VintageData(
        "USD_INDX_SA",
        release_lags=[3, 20, 25],
        number_firsts=2,
        shortest=8,
        freq="Q",
        seasonal=10,
        added_dates=4,
    )
    dfq1 = vins_q.make_grade1()
    dfq1.groupby("release_date").agg(["mean", "count"])
    dfq2 = vins_q.make_grade2()

    vins_w = VintageData(
        "USD_INDX_SA",
        cutoff="2019-06-30",
        release_lags=[3, 20, 25],
        number_firsts=3 * 52,
        shortest=26,
        freq="W",
        seasonal=10,
        added_dates=52,
    )
    dfw1 = vins_w.make_grade1()
    dfw1.groupby("release_date").agg(["mean", "count"])
    dfw2 = vins_w.make_grade2()

    dfm1.info()



"""
Module with functions for processing "blacklist" data for cross-sections in a
quantamental DataFrame.
"""

import numpy as np
import pandas as pd
from typing import List
from itertools import groupby
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.simulate import make_qdf_black, make_qdf
from macrosynergy.management.types import QuantamentalDataFrame


def make_blacklist(
    df: QuantamentalDataFrame,
    xcat: str,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    nan_black: bool = False,
):
    """
    Converts binary category of standardized dataframe into a standardized dictionary
    that can serve as a blacklist for cross-sections in further analyses

    Parameters
    ----------
    df : QuantamentalDataFrame
        standardized DataFrame with following columns: 'cid', 'xcat', 'real_date' and
        'value'.
    xcat : str
        category with binary values, where 1 means blacklisting and 0 means not
        blacklisting.
    cids : str
        list of cross-sections that are considered in the formation of the blacklist.
        Per default, all available cross sections are considered.
    start : str
        earliest date in ISO format. Default is None and earliest date for which the
        respective category is available is used.
    end : str
        latest date in ISO format. Default is None and latest date for which the
        respective category is available is used.
    nan_black : bool
        if True NaNs are blacklisted (coverted to ones). Defaults is False, i.e. NaNs
        are converted to zeroes.

    Returns
    -------
    dict
        standardized dictionary with cross-sections as keys and tuples of start and end
        dates of the blacklist periods in ISO formats as values. If one cross section has
        multiple blacklist periods, numbers are added to the keys (i.e. TRY_1, TRY_2, etc.)
    """

    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("df must be a standardized quantamental dataframe")

    df = QuantamentalDataFrame(df)

    dfd = reduce_df(df=df, xcats=[xcat], cids=cids, start=start, end=end)

    if "value" not in dfd.columns:
        raise ValueError("`value` column not found in df")

    if not all(np.isin(dfd["value"].dropna().unique(), [0, 1])):
        raise ValueError("blacklist values must all be 0/1")

    df_pivot = dfd.pivot(index="real_date", columns="cid", values="value")
    dates = df_pivot.index
    cids_df = list(df_pivot.columns)

    # replace NaNs
    df_pivot[df_pivot.isna()] = int(nan_black)  # 1 if nan_black else 0

    dates_dict = {}
    for cid in cids_df:
        count = 0
        column = df_pivot[cid].to_numpy()
        si = 0
        for k, v in groupby(column):  # iterator of consecutive keys and values
            v = list(v)  # instantiate the iterable in memory.
            length = len(v)
            if v[0] == 1:  # if blacklist period
                if count == 0:
                    dates_dict[cid] = startend(dates, si, length)
                elif count == 1:
                    val = dates_dict.pop(cid)
                    dates_dict[cid + "_1"] = val  # change key if more than 1 per cid
                    count += 1
                    dates_dict[cid + "_" + str(count)] = startend(dates, si, length)
                else:
                    dates_dict[cid + "_" + str(count)] = startend(dates, si, length)
                count += 1
            si += length
    return dates_dict


def startend(dti, start, length):
    """
    Return start and end dates of a sequence as tuple

    Parameters
    ----------
    dti : DateTimeIndex
        datetime series of working days
    start : int
        index of start
    length : int
        number of sequential days

    Returns
    -------
    Tuple[pd.Timestamp, pd.Timestamp]
        tuple of start and end date
    """

    tup = (dti[start], dti[start + (length - 1)])
    return tup


if __name__ == "__main__":
    cids = ["AUD", "GBP", "CAD", "USD"]
    cols = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cid1 = pd.DataFrame(index=cids, columns=cols)

    df_cid1.loc["AUD"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cid1.loc["GBP"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cid1.loc["CAD"] = ["2011-01-01", "2021-11-30", 0, 1]
    df_cid1.loc["USD"] = ["2011-01-01", "2020-12-30", 0, 1]

    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]
    df_xcat1 = pd.DataFrame(index=["FXXR_NSA", "FXCRY_NSA"], columns=cols)
    df_xcat1.loc["FXXR_NSA"] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcat1.loc["FXCRY_NSA"] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df1 = make_qdf(df_cid1, df_xcat1, back_ar=0.05)

    df_xcat2 = pd.DataFrame(index=["FXNONTRADE_NSA"], columns=["earliest", "latest"])
    df_xcat2.loc["FXNONTRADE_NSA"] = ["2010-01-01", "2021-11-30"]
    black = {
        "AUD": ("2010-01-12", "2010-06-14"),
        "USD": ("2011-08-17", "2011-09-20"),
        "CAD_1": ("2011-01-04", "2011-01-23"),
        "CAD_2": ("2013-01-09", "2013-04-10"),
        "CAD_3": ("2015-01-12", "2015-03-12"),
        "CAD_4": ("2021-11-01", "2021-11-20"),
    }

    print(black)
    df2 = make_qdf_black(df_cid1, df_xcat2, blackout=black)

    df = pd.concat([df1, df2]).reset_index()

    dates_dict = make_blacklist(
        df, xcat="FXNONTRADE_NSA", cids=None, start=None, end=None
    )

    # If the output, from the below printed dictionary, differs from the above defined
    # dictionary, it should be by a date or two, as the construction of the dataframe,
    # using make_qdf_black(), will account for the dates received, in the dictionary,
    # being weekends. Therefore, if any of the dates, for the start or end of the
    # blackout period are Saturday or Sunday, the date for will be shifted to the
    # following Monday. Hence, a break in alignment from "blackout" to "dates_dict".
    print(dates_dict)



from .lad_regressor import (
    LADRegressor,
)

from .weighted_lad_regressors import (
    SignWeightedLADRegressor,
    TimeWeightedLADRegressor,
)

__all__ = [
    "LADRegressor",
    "SignWeightedLADRegressor",
    "TimeWeightedLADRegressor",
]


"""
Module hosting custom types and meta-classes for use with Quantamental DataFrames.
"""

from typing import List, Optional, Any, Iterable, Mapping, Union, Dict, Set, Tuple
import pandas as pd
import numpy as np
import warnings
import itertools
import functools

from macrosynergy.management.constants import JPMAQS_METRICS

from .base import QuantamentalDataFrameBase


def get_col_sort_order(df: QuantamentalDataFrameBase) -> List[str]:
    """
    Sort the columns of a QuantamentalDataFrame (in-place) in a consistent order.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to return the sorted columns of.

    Returns
    -------
    List[str]
        List of sorted column names.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    metric_cols: Set = set(df.columns) - set(QuantamentalDataFrameBase.IndexCols)
    non_jpmaqs_metrics: List[str] = sorted(metric_cols - set(JPMAQS_METRICS))
    jpmaqs_metrics: List[str] = [m for m in JPMAQS_METRICS if m in metric_cols]

    return QuantamentalDataFrameBase.IndexCols + jpmaqs_metrics + non_jpmaqs_metrics


def change_column_format(
    df: QuantamentalDataFrameBase,
    cols: List[str],
    dtype: Any,
) -> QuantamentalDataFrameBase:
    """
    Change the format of columns in a DataFrame.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to change the format of.
    cols : List[str]
        List of column names to change the format of.
    dtype : Any
        Data type to change the columns to.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the columns changed to the specified format

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.

    TypeError
        If `cols` is not a list of strings.

    ValueError
        If a column in `cols` is not found in the DataFrame.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a pandas DataFrame.")

    if not isinstance(cols, list) or not all([isinstance(col, str) for col in cols]):
        raise TypeError("`cols` must be a list of strings.")

    for col in cols:
        try:
            if col not in df.columns:
                raise ValueError(f"Column '{col}' not found in DataFrame.")
            if not df[col].dtype == dtype:
                df[col] = df[col].astype(dtype)
        except Exception as exc:
            warnings.warn(f"Could not convert column {col} to {dtype}. Error: {exc}")

    return df


def qdf_to_categorical(
    df: QuantamentalDataFrameBase,
) -> QuantamentalDataFrameBase:
    """
    Convert the index columns ("cid", "xcat") of a DataFrame to categorical format.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to convert the index columns of.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the index columns converted to categorical format.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    df = change_column_format(df, QuantamentalDataFrameBase._StrIndexCols, "category")
    return df


def qdf_to_string_index(
    df: QuantamentalDataFrameBase,
) -> QuantamentalDataFrameBase:
    """
    Convert the index columns ("cid", "xcat") of a DataFrame to string format.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to convert the index columns of.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the index columns converted to string format.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    df = change_column_format(df, QuantamentalDataFrameBase._StrIndexCols, "object")
    return df


def check_is_categorical(df: QuantamentalDataFrameBase) -> bool:
    """
    Check if the index columns of a DataFrame are categorical.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to check the index columns of.

    Returns
    -------
    bool
        True if the required index columns ("cid", "xcat") are categorical, False
        otherwise.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    return all(
        df[col].dtype.name == "category"
        for col in QuantamentalDataFrameBase._StrIndexCols
    )


def _get_tickers_series(
    df: QuantamentalDataFrameBase,
    cid_column: str = "cid",
    xcat_column: str = "xcat",
) -> pd.Categorical:
    """
    Get the list of tickers from the DataFrame.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to extract the tickers from.

    cid_column : str, optional
        Column name of the `cid` in the DataFrame. Default is "cid".

    xcat_column : str, optional
        Column name of the `xcat` in the DataFrame. Default is "xcat".
    """
    # check if the columns are in the dataframe and are categorical
    if cid_column not in df.columns:
        raise ValueError(f"Column '{cid_column}' not found in DataFrame.")
    if xcat_column not in df.columns:
        raise ValueError(f"Column '{xcat_column}' not found in DataFrame.")

    if not check_is_categorical(df):
        return df[cid_column] + "_" + df[xcat_column]

    cid_labels = df["cid"].cat.categories[df["cid"].cat.codes]
    xcat_labels = df["xcat"].cat.categories[df["xcat"].cat.codes]

    ticker_labels = [f"{cid}_{xcat}" for cid, xcat in zip(cid_labels, xcat_labels)]
    categories = pd.unique(pd.Series(ticker_labels))

    ticker_series = pd.Categorical(
        ticker_labels,
        categories=categories,
        ordered=True,
    )

    return ticker_series


def apply_blacklist(
    df: QuantamentalDataFrameBase,
    blacklist: Mapping[str, Iterable[Union[str, pd.Timestamp]]],
) -> QuantamentalDataFrameBase:
    """
    Apply a blacklist to a list of `cids` and `xcats`. The blacklisted data ranges are
    removed from the DataFrame. This is useful for removing data that is known to be
    incorrect or unreliable.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to apply the blacklist to.

    blacklist : dict
        Dictionary with keys as `cids` and values as a list of start and end dates
        to blacklist. Example:

        .. code-block:: python

            {"cid": ["2020-01-01", "2020-12-31"]}

        This can be extended to cover multiple periods for the same `cid` by appending
        an additional label to the end of the `cid` key. Example:

        .. code-block:: python

            {
                "usd_1": ["2020-01-01", "2020-12-31"],
                "usd_2": ["2020-01-01", "2020-12-31"],
                "eur": ["2020-01-01", "2020-12-31"],
            }

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the blacklist applied.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a pandas DataFrame.")

    if not isinstance(blacklist, dict):
        raise TypeError("`blacklist` must be a dictionary.")

    if not all([isinstance(k, str) for k in blacklist.keys()]):
        raise TypeError("Keys of `blacklist` must be strings.")

    if not all([isinstance(v, Iterable) for v in blacklist.values()]):
        raise TypeError("Values of `blacklist` must be iterables.")

    if not all(
        [isinstance(vv, (str, pd.Timestamp)) for v in blacklist.values() for vv in v]
    ) or any([len(v) != 2 for v in blacklist.values()]):
        raise TypeError(
            "Values of `blacklist` must be lists of start & end dates (str or pd.Timestamp)."
        )

    for key, value in blacklist.items():
        df = df[
            ~(
                (df["cid"] == key[:3])
                & (df["real_date"] >= value[0])
                & (df["real_date"] <= value[1])
            )
        ]

    return df.reset_index(drop=True)


def _sync_df_categories(
    df: QuantamentalDataFrameBase,
) -> QuantamentalDataFrameBase:
    """
    Sync the categories of the DataFrame with the data.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to sync the categories of.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the categories synced.
    """
    if not check_is_categorical(df):
        return df

    df["cid"] = df["cid"].cat.remove_unused_categories().astype("category")
    df["xcat"] = df["xcat"].cat.remove_unused_categories().astype("category")

    return df


def reduce_df(
    df: QuantamentalDataFrameBase,
    cids: Optional[List[str]] = None,
    xcats: Optional[List[str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: dict = None,
    out_all: bool = False,
    intersect: bool = False,
) -> Union[
    QuantamentalDataFrameBase, Tuple[QuantamentalDataFrameBase, List[str], List[str]]
]:
    """
    Filter DataFrame by `cids`, `xcats`, and `start` & `end` dates.

    Parameters
    ----------
    df : QuantamentalDataFrameBase
        The DataFrame to be filtered.
    cids : Optional[List[str]], optional
        List of `cid` values to filter by. If None, all `cid` values are included.
    xcats : Optional[List[str]], optional
        List of `xcat` values to filter by. If None, all `xcat` values are included.
    start : Optional[str], optional
        Start date for filtering. If None, no start date filtering is applied.
    end : Optional[str], optional
        End date for filtering. If None, no end date filtering is applied.
    blacklist : dict, optional
        Dictionary specifying blacklist criteria. If None, no blacklist filtering is
        applied.
    out_all : bool, optional
        If True, returns the filtered DataFrame along with the lists of `xcats` and
        `cids`; i.e. `(df, xcats, cids)`.
    intersect : bool, optional
        If True, only includes `cid` values that are present for all `xcat` values.

    Returns
    -------
    Union[QuantamentalDataFrameBase, Tuple[QuantamentalDataFrameBase, List[str], List[str]]]
        The filtered DataFrame. If `out_all` is True, also returns the lists of `xcats`
        and `cids`.
    """
    if xcats is not None:
        if not isinstance(xcats, list):
            xcats = [xcats]

    if start:
        df = df[df["real_date"] >= pd.to_datetime(start)]

    if end:
        df = df[df["real_date"] <= pd.to_datetime(end)]

    if blacklist is not None:
        df = apply_blacklist(df, blacklist)

    if xcats is None:
        xcats = sorted(df["xcat"].unique())
    else:
        xcats_in_df = df["xcat"].unique()
        xcats = [xcat for xcat in xcats if xcat in xcats_in_df]

    df = df[df["xcat"].isin(xcats)]

    if intersect:
        cids_in_df = set.intersection(
            *(set(df[df["xcat"] == xcat]["cid"].unique()) for xcat in xcats)
        )
    else:
        cids_in_df = df["cid"].unique()

    if cids is None:
        cids = sorted(cids_in_df)
    else:
        cids = [cids] if isinstance(cids, str) else cids
        cids = [cid for cid in cids if cid in cids_in_df]

    df = df[df["cid"].isin(cids)].reset_index(drop=True)

    df = _sync_df_categories(df)

    df = df.drop_duplicates().reset_index(drop=True)

    if out_all:
        return df, xcats, sorted(cids)
    else:
        return df


def reduce_df_by_ticker(
    df: QuantamentalDataFrameBase,
    tickers: List[str],
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: dict = None,
) -> QuantamentalDataFrameBase:
    """
    Filters the given QuantamentalDataFrameBase based on tickers, date range, and
    blacklist.

    Parameters
    ----------
    df : QuantamentalDataFrameBase
        The DataFrame to be filtered.
    tickers : List[str]
        List of tickers to filter by.
    start : Optional[str], optional
        Start date for filtering. If None, no start date filtering is applied.
    end : Optional[str], optional
        End date for filtering. If None, no end date filtering is applied.
    blacklist : dict, optional
        Dictionary specifying blacklist criteria. If None, no blacklist filtering is
        applied.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.

    Returns
    -------
    QuantamentalDataFrameBase
        The filtered DataFrame.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    if not isinstance(tickers, list):
        if tickers is not None:
            raise TypeError("`tickers` must be a list of strings.")

    if start is not None:
        df = df.loc[df["real_date"] >= pd.to_datetime(start)]
    if end is not None:
        df = df.loc[df["real_date"] <= pd.to_datetime(end)]

    if blacklist is not None:
        df = apply_blacklist(df, blacklist)

    ticker_series = _get_tickers_series(df)
    if tickers is None:
        tickers = sorted(ticker_series.unique())

    df = df[ticker_series.isin(tickers)].reset_index(drop=True)

    df = _sync_df_categories(df)

    return df.drop_duplicates().reset_index(drop=True)


def update_df(
    df: QuantamentalDataFrameBase,
    df_add: QuantamentalDataFrameBase,
    xcat_replace: bool = False,
) -> QuantamentalDataFrameBase:
    """
    Append a standard DataFrame to a standard base DataFrame with ticker replacement on
    the intersection.

    Parameters
    ----------
    df : QuantamentalDataFrame
        Base DataFrame to append to.
    df_add : QuantamentalDataFrame
        DataFrame to append.
    xcat_replace : bool, optional
        If True, replace the xcats in the base DataFrame with the xcats in the DataFrame
        to append. Default is False.
    """

    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")
    if not isinstance(df_add, QuantamentalDataFrameBase):
        raise TypeError("`df_add` must be a QuantamentalDataFrame.")
    if not isinstance(xcat_replace, bool):
        raise TypeError("`xcat_replace` must be a boolean.")

    if xcat_replace:
        df = update_categories(df=df, df_add=df_add)
    else:
        df = update_tickers(df=df, df_add=df_add)
    _sortorder = QuantamentalDataFrameBase.IndexColsSortOrder
    return df.sort_values(_sortorder).reset_index(drop=True)


def update_tickers(
    df: pd.DataFrame,
    df_add: pd.DataFrame,
) -> QuantamentalDataFrameBase:
    """
    Method used to update aggregate DataFrame on the ticker level.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to update.
    df_add : pd.DataFrame
        DataFrame to add to the base DataFrame.

    Returns
    -------
    QuantamentalDataFrame
        Updated DataFrame.
    """
    if df_add.empty:
        return df
    elif df.empty:
        return df_add

    if all(
        _df[icol].dtype.name == "category"
        for _df in [df, df_add]
        for icol in ["cid", "xcat"]
    ):
        union_cids = pd.api.types.union_categoricals(
            [df["cid"].unique(), df_add["cid"].unique()]
        )
        union_xcats = pd.api.types.union_categoricals(
            [df["xcat"].unique(), df_add["xcat"].unique()]
        )
        df["cid"] = pd.Categorical(df["cid"], categories=union_cids.categories)
        df["xcat"] = pd.Categorical(df["xcat"], categories=union_xcats.categories)

        df_add["cid"] = pd.Categorical(df_add["cid"], categories=union_cids.categories)
        df_add["xcat"] = pd.Categorical(
            df_add["xcat"], categories=union_xcats.categories
        )

    df = pd.concat([df, df_add], axis=0, ignore_index=True)

    df = df.drop_duplicates(
        subset=QuantamentalDataFrameBase.IndexCols,
        keep="last",
    ).reset_index(drop=True)
    return df


def update_categories(
    df: QuantamentalDataFrameBase,
    df_add: QuantamentalDataFrameBase,
) -> QuantamentalDataFrameBase:
    """
    Method used to update the DataFrame on the category level.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")
    if not isinstance(df_add, QuantamentalDataFrameBase):
        raise TypeError("`df_add` must be a QuantamentalDataFrame.")

    incumbent_categories = list(df["xcat"].unique())
    new_categories = list(df_add["xcat"].unique())

    append_condition = set(incumbent_categories) | set(new_categories)
    intersect = list(set(incumbent_categories).intersection(set(new_categories)))

    if all(
        _df[icol].dtype.name == "category"
        for _df in [df, df_add]
        for icol in ["cid", "xcat"]
    ):
        union_cids = pd.api.types.union_categoricals(
            [df["cid"].unique(), df_add["cid"].unique()]
        )
        union_xcats = pd.api.types.union_categoricals(
            [df["xcat"].unique(), df_add["xcat"].unique()]
        )
        df["cid"] = pd.Categorical(df["cid"], categories=union_cids.categories)
        df["xcat"] = pd.Categorical(df["xcat"], categories=union_xcats.categories)

        df_add["cid"] = pd.Categorical(df_add["cid"], categories=union_cids.categories)
        df_add["xcat"] = pd.Categorical(
            df_add["xcat"], categories=union_xcats.categories
        )

    if len(append_condition) == len(incumbent_categories + new_categories):
        df = pd.concat([df, df_add], axis=0, ignore_index=True)

    else:
        df = df[~df["xcat"].isin(intersect)]
        df = pd.concat([df, df_add], axis=0, ignore_index=True)
    _sortorder = QuantamentalDataFrameBase.IndexColsSortOrder
    return df.sort_values(_sortorder).reset_index(drop=True)


def qdf_to_wide_df(
    df: QuantamentalDataFrameBase,
    value_column: str = "value",
) -> pd.DataFrame:
    """
    Pivot the DataFrame to a wide format with memory efficiency.
    """
    # Ensure inputs are of the correct type and exist in the DataFrame
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")
    if not isinstance(value_column, str):
        raise TypeError("`value_column` must be a string.")
    if value_column not in df.columns:
        raise ValueError(f"Column '{value_column}' not found in DataFrame.")

    df["ticker"] = _get_tickers_series(df)

    # Perform the pivot directly within the assignment to reduce memory footprint
    return df.pivot(
        index="real_date", columns="ticker", values=value_column
    ).rename_axis(None, axis=1)


def add_ticker_column(
    df: QuantamentalDataFrameBase,
) -> List[str]:
    """
    Get the list of tickers from the DataFrame.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to extract the tickers from.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.

    Returns
    -------
    List[str]
        List of tickers.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a pandas DataFrame.")

    df["ticker"] = _get_tickers_series(df)

    return df


def _add_categorical_column(
    df: pd.DataFrame,
    column_name: str,
    fill_value: str,
) -> pd.DataFrame:
    """
    Add a categorical index column to a DataFrame. Typically `cid` or `xcat`.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to add the index column to.
    column_name : str
        Name of the index column to add.
    """
    df[column_name] = pd.Categorical.from_codes([0] * len(df), categories=[fill_value])
    return df


def rename_xcats(
    df: QuantamentalDataFrameBase,
    xcat_map: Optional[Dict[str, str]] = None,
    select_xcats: Optional[List[str]] = None,
    postfix: Optional[str] = None,
    prefix: Optional[str] = None,
    name_all: Optional[str] = None,
    fmt_string: Optional[str] = None,
) -> QuantamentalDataFrameBase:
    """
    Rename the xcats in a DataFrame based on a mapping or a format string. Only one of
    `xcat_map` or `select_xcats` must be provided. If `name_all` is provided, all xcats
    will be renamed to this value.

    NOTE: This function maintains the datatype of the xcat column as a categorical.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to rename the xcats in.
    xcat_map : dict, optional
        Dictionary mapping the old xcats to new xcats. Default is None.
    select_xcats : List[str], optional
        List of xcats to rename. Default is None.
    postfix : str, optional
        Postfix to add to the xcats. Default is None.
    prefix : str, optional
        Prefix to add to the xcats. Default is None.
    name_all : str, optional
        Name to rename all xcats to. Default is None.
    fmt_string : str, optional
        Format string to rename xcats. Default is None.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.
    ValueError
        If both `xcat_map` and `select_xcats` are provided.
    TypeError
        If `xcat_map` is not a dictionary with string keys and values.
    ValueError
        If `postfix`, `prefix`, `name_all`, or `fmt_string` are not provided.
    ValueError
        If `fmt_string` does not contain exactly one pair of curly braces.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the xcats renamed.
    """
    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    if bool(xcat_map) and bool(select_xcats):
        raise ValueError("Only one of `xcat_map` or `select_xcats` must be provided.")

    # Validate `xcat_map`
    if xcat_map is not None:
        if not (
            isinstance(xcat_map, dict)
            and all(
                isinstance(k, str) and isinstance(v, str) for k, v in xcat_map.items()
            )
        ):
            raise TypeError(
                "`xcat_map` must be a dictionary with string keys and values."
            )
        # Rename xcats based on `xcat_map`
        df["xcat"] = df["xcat"].cat.rename_categories(
            {old_cat: xcat_map.get(old_cat, old_cat) for old_cat in df["xcat"].unique()}
        )
        return df

    if select_xcats is None:
        select_xcats = df["xcat"].unique()

    # Ensure exactly one of postfix, prefix, name_all, or fmt_string is provided
    if not (bool(postfix) ^ bool(prefix) ^ bool(name_all) ^ bool(fmt_string)):
        raise ValueError(
            "Exactly one of `postfix`, `prefix`, `name_all`, or `fmt_string` must be provided."
        )

    funcs = {
        "postfix": lambda x: f"{x}{postfix}",
        "prefix": lambda x: f"{prefix}{x}",
        "name_all": lambda x: name_all,
        "fmt_string": lambda x: fmt_string.format(x),
    }

    curr_func = None
    for var_, name_ in zip(
        [postfix, prefix, name_all, fmt_string],
        ["postfix", "prefix", "name_all", "fmt_string"],
    ):
        if var_ is not None:
            curr_func = name_

    if fmt_string is not None:
        if fmt_string.count("{}") != 1:
            raise ValueError(
                "The `fmt_string` must contain exactly one pair of curly braces."
            )

    if name_all is not None:
        xc_col = df["xcat"].astype(str)
        xc_col = xc_col.replace({cat: funcs[curr_func](cat) for cat in select_xcats})
        df["xcat"] = pd.Categorical(xc_col, categories=list(set(xc_col)))

    else:
        df["xcat"] = df["xcat"].cat.rename_categories(
            {cat: funcs[curr_func](cat) for cat in select_xcats}
        )

    return df


def create_empty_categorical_qdf(
    cid: Optional[str] = None,
    xcat: Optional[str] = None,
    ticker: Optional[str] = None,
    metrics: List[str] = ["value"],
    date_range: Optional[pd.DatetimeIndex] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    categorical: bool = True,
) -> QuantamentalDataFrameBase:
    """
    Create an empty QuantamentalDataFrame with categorical index columns. This is useful
    for creating a DataFrame for a given ticker with the required metrics.
    The ticker can be specified using `cid` and `xcat` or directly using `ticker`. The
    data range can be specified using `date_range` or `start` and `end`.

    Parameters
    ----------
    cid : str, optional
        `cid` value to use. Must be passed with `xcat`. Default is None.
    xcat : str, optional
        `xcat` value to use. Must be passed with `cid`. Default is None.
    ticker : str, optional
        Ticker to use. Must not be passed with `cid` and `xcat`. Default is None.
    metrics : List[str], optional
        List of metrics to create columns for. Default is ["value"].
    date_range : pd.DatetimeIndex, optional
        Date range to create the DataFrame for. Must not be passed with `start` and `end`.
        Default is None.
    start : str, optional
        Start date for the DataFrame. Default is None.
    end : str, optional
        End date for the DataFrame. Default is None.

    Raises
    ------
    TypeError
        If `metrics` is not a list of strings.
    ValueError
        If `date_range` is None and `start` and `end` are not provided.
    ValueError
        If `cid` and `xcat` are not provided together.
    ValueError
        If `cid` and `xcat` are provided together.
    ValueError
        If `ticker` is provided with `cid` and `xcat`.

    Returns
    -------
    QuantamentalDataFrame
        Empty DataFrame with the required index columns and metrics.
    """
    if not all(isinstance(m, str) for m in metrics):
        raise TypeError("`metrics` must be a list of strings.")

    if (date_range is None) and (start is None or end is None):
        raise ValueError(
            "Either `date_range` or `start_date` & `end_date` must be specified."
        )

    if date_range is None:
        date_range = pd.bdate_range(start=start, end=end)

    if bool(cid) ^ bool(xcat):
        raise ValueError("`cid` and `xcat` must be specified together.")

    if not (bool(cid) ^ bool(ticker)):
        raise ValueError("Either specify `cid` & `xcat` or `ticker` but not both.")

    if ticker is not None:
        cid, xcat = ticker.split("_", 1)

    qdf = pd.DataFrame(columns=["real_date"], data=date_range)
    qdf = _add_categorical_column(qdf, "cid", cid)
    qdf = _add_categorical_column(qdf, "xcat", xcat)

    for metric in metrics:
        qdf[metric] = np.nan

    if not categorical:
        qdf = qdf_to_string_index(qdf)

    return qdf


def add_nan_series(
    df: QuantamentalDataFrameBase,
    ticker: Optional[str] = None,
    cid: Optional[str] = None,
    xcat: Optional[str] = None,
    start: Optional[Union[str, pd.Timestamp]] = None,
    end: Optional[Union[str, pd.Timestamp]] = None,
) -> QuantamentalDataFrameBase:
    """
    Add a NaN series to the DataFrame for a given ticker.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to add the NaN series to.
    ticker : str, optional
        Ticker to add the NaN series for. Must not be passed with `cid` and `xcat`.
        Default is None.
    cid : str, optional
        `cid` value to use. Must be passed with `xcat`. Default is None.
    xcat : str, optional
        `xcat` value to use. Must be passed with `cid`. Default is None.
    start : str or pd.Timestamp, optional
        Start date for the NaN series. Default is None.
    end : str or pd.Timestamp, optional
        End date for the NaN series. Default is None.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.
    ValueError
        If `ticker` is provided with `cid` and `xcat`.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the NaN series added.
    """

    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("`df` must be a QuantamentalDataFrame.")

    metrics = df.columns.difference(QuantamentalDataFrameBase.IndexCols)

    if start is None:
        start = df["real_date"].min()
    if end is None:
        end = df["real_date"].max()

    nan_df = create_empty_categorical_qdf(
        cid=cid,
        xcat=xcat,
        ticker=ticker,
        start=start,
        end=end,
        metrics=metrics,
        categorical=check_is_categorical(df),
    )

    df = update_df(df=df, df_add=nan_df)
    return df


def drop_nan_series(
    df: QuantamentalDataFrameBase, column: str = "value", raise_warning: bool = False
) -> QuantamentalDataFrameBase:
    """
    Drops any series that are entirely NaNs. Raises a user warning if any series are
    dropped.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame to drop the NaN series from.
    column : str, optional
        Column to check for NaNs. Default is "value".
    raise_warning : bool, optional
        If True, raises a warning if any series are dropped. Default is False.

    Raises
    ------
    TypeError
        If `df` is not a QuantamentalDataFrame.
    ValueError
        If `column` is not found in the DataFrame.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the NaN series dropped.
    """

    if not isinstance(df, QuantamentalDataFrameBase):
        raise TypeError("Argument `df` must be a Quantamental DataFrame.")

    if column not in df.columns:
        raise ValueError(f"Column {column} not present in DataFrame.")

    if not df[column].isna().any():
        return df

    if not isinstance(raise_warning, bool):
        raise TypeError("Error: The raise_warning argument must be a boolean.")

    df_orig: pd.DataFrame = df.copy()
    for cd, xc in df_orig.groupby(["cid", "xcat"], observed=True).groups:
        sel_series: pd.Series = df_orig[
            (df_orig["cid"] == cd) & (df_orig["xcat"] == xc)
        ][column]
        if sel_series.isna().all():
            if raise_warning:
                warnings.warn(
                    message=f"The series {cd}_{xc} is populated "
                    "with NaNs only, and will be dropped.",
                    category=UserWarning,
                )
            df = df[~((df["cid"] == cd) & (df["xcat"] == xc))]

    return df.reset_index(drop=True)


def qdf_from_timeseries(
    timeseries: pd.Series,
    cid: Optional[str] = None,
    xcat: Optional[str] = None,
    ticker: Optional[str] = None,
    metric: str = "value",
) -> QuantamentalDataFrameBase:
    """
    Create a QuantamentalDataFrame from a time series.

    Parameters
    ----------
    timeseries : pd.Series
        Time series to create the QuantamentalDataFrame from.
    cid : str, optional
        `cid` value to use. Must be passed with `xcat`. Default is None.
    xcat : str, optional
        `xcat` value to use. Must be passed with `cid`. Default is None.
    ticker : str, optional
        Ticker to use. Must not be passed with `cid` and `xcat`. Default is None.
    metric : str, optional
        Metric name to use. Default is "value".

    Raises
    ------
    TypeError
        If `timeseries` is not a pandas Series.
    TypeError
        If `metric` is not a string.
    ValueError
        If `timeseries` does not have a datetime index.
    ValueError
        If only one of `cid` and `xcat` is provided.
    ValueError
        If `ticker` is provided with `cid` and `xcat`.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame created from the time series.
    """
    if not isinstance(timeseries, pd.Series):
        raise TypeError("`timeseries` must be a pandas Series.")
    if not isinstance(metric, str):
        raise TypeError("`metric` must be a string.")

    if not isinstance(timeseries.index, pd.DatetimeIndex):
        raise ValueError("`timeseries` must have a datetime index.")

    if (cid is None) ^ (xcat is None):
        raise ValueError("Both `cid` and `xcat` must be provided.")

    if not ((cid is None) ^ (ticker is None)):
        raise ValueError("Either provide `cid` & `xcat` or `ticker`.")

    if ticker is not None:
        cid, xcat = ticker.split("_", 1)

    assert bool(cid) and bool(xcat)

    df = timeseries.reset_index().rename(columns={"index": "real_date", 0: metric})
    # assign as categorical string
    df = _add_categorical_column(df, "cid", cid)
    df = _add_categorical_column(df, "xcat", xcat)

    df = df[[*QuantamentalDataFrameBase.IndexCols, metric]]
    return QuantamentalDataFrameBase(df)


def _convert_to_single_metric_qdfs(
    qdf: QuantamentalDataFrameBase,
) -> QuantamentalDataFrameBase:
    """
    Internal function for concat_qdfs. Splits a QuantamentalDataFrame with multiple
    metrics to multiple QuantamentalDataFrames with a single metric.

    Parameters
    ----------
    qdf : QuantamentalDataFrame
        DataFrame to convert.

    Returns
    -------
    List[QuantamentalDataFrame]
        List of QuantamentalDataFrames with a single metric.
    """
    return [
        qdf[[*QuantamentalDataFrameBase.IndexCols, metric]]
        for metric in qdf.columns.difference(QuantamentalDataFrameBase.IndexCols)
    ]


def concat_qdfs(
    qdf_list: List[QuantamentalDataFrameBase],
) -> QuantamentalDataFrameBase:
    """
    Concatenate a list of QuantamentalDataFrames into a single QuantamentalDataFrame.
    Converts the index columns to categorical format, if not already categorical.

    Parameters
    ----------
    qdf_list : List[QuantamentalDataFrame]
        List of QuantamentalDataFrames to concatenate.

    Raises
    ------
    TypeError
        If `qdf_list` is not a list of QuantamentalDataFrames.

    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the QuantamentalDataFrames concatenated.
    """
    if not isinstance(qdf_list, list):
        raise TypeError("`qdfs_list` must be a list of QuantamentalDataFrames.")

    if not all(isinstance(qdf, QuantamentalDataFrameBase) for qdf in qdf_list):
        raise TypeError("All elements in `qdfs_list` must be QuantamentalDataFrames.")

    if len(qdf_list) == 0:
        raise ValueError("`qdfs_list` is empty.")

    for iq, qdf in enumerate(qdf_list):
        qdf_list[iq] = qdf_to_categorical(qdf)

    comb_cids = pd.api.types.union_categoricals(
        [qdf["cid"].unique() for qdf in qdf_list]
    )
    comb_xcats = pd.api.types.union_categoricals(
        [qdf["xcat"].unique() for qdf in qdf_list]
    )

    for iq, qdf in enumerate(qdf_list):
        qdf_list[iq]["cid"] = pd.Categorical(
            qdf["cid"], categories=comb_cids.categories
        )
        qdf_list[iq]["xcat"] = pd.Categorical(
            qdf["xcat"], categories=comb_xcats.categories
        )

    qdf_list = list(itertools.chain(*map(_convert_to_single_metric_qdfs, qdf_list)))

    def _get_metric(df: QuantamentalDataFrameBase) -> str:
        return list(set(df.columns) - set(QuantamentalDataFrameBase.IndexCols))[0]

    def _group_by_metric(
        dfl: List[QuantamentalDataFrameBase], fm: List[str]
    ) -> List[List[QuantamentalDataFrameBase]]:
        r = [[] for _ in range(len(fm))]
        while dfl:
            metric = _get_metric(df=dfl[0])
            r[fm.index(metric)] += [dfl.pop(0)]
        return r

    found_metrics = list(set(map(_get_metric, qdf_list)))
    qdf_list = _group_by_metric(dfl=qdf_list, fm=found_metrics)

    df: pd.DataFrame = functools.reduce(
        lambda left, right: pd.merge(
            left,
            right,
            on=QuantamentalDataFrameBase.IndexCols,
            how="outer",
        ),
        map(
            lambda x: pd.concat(qdf_list.pop(0), axis=0, ignore_index=False),
            found_metrics,
        ),
    )

    return df.sort_values(by=QuantamentalDataFrameBase.IndexColsSortOrder).reset_index(
        drop=True
    )[get_col_sort_order(df)]



"""
A subclass inheriting from `macrosynergy.visuals.plotter.Plotter`, designed to plot time
series data as a heatmap.
"""

from numbers import Number
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from seaborn.utils import relative_luminance

from macrosynergy.management.simulate import make_test_df
from macrosynergy.management.utils import downsample_df_on_real_date
from macrosynergy.visuals.plotter import Plotter


class Heatmap(Plotter):
    """
    Class for plotting time series data as a heatmap. Inherits from
    `macrosynergy.visuals.plotter.Plotter`.

    Parameters
    ----------
    df : ~pandas.DataFrame
        A DataFrame with the following columns: 'cid', 'xcat', 'real_date', and at least
        one metric from - 'value', 'grading', 'eop_lag', or 'mop_lag'.
    cids : List[str]
        A list of cids to select from the DataFrame. If None, all cids are selected.
    xcats : List[str]
        A list of xcats to select from the DataFrame. If None, all xcats are selected.
    metrics : List[str]
        A list of metrics to select from the DataFrame. If None, all metrics are
        selected.
    start : str
        ISO-8601 formatted date. Select data from this date onwards. If None, all dates
        are selected.
    end : str
        ISO-8601 formatted date. Select data up to and including this date. If None, all
        dates are selected.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            start=start,
            end=end,
            *args,
            **kwargs,
        )

    def plot(
        self,
        df: pd.DataFrame,
        figsize: Tuple[Number, Number] = (12, 8),
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 14,
        title: Optional[str] = None,
        title_fontsize: int = 22,
        title_xadjust: Number = 0.5,
        title_yadjust: Number = 1.0,
        vmin: Optional[Number] = None,
        vmax: Optional[Number] = None,
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        on_axis: Optional[plt.Axes] = None,
        max_xticks: int = 50,
        cmap: Optional[Union[str, mpl.colors.Colormap]] = None,
        rotate_xticks: Optional[Number] = 0,
        rotate_yticks: Optional[Number] = 0,
        show_tick_lines: Optional[bool] = True,
        show_colorbar: Optional[bool] = True,
        show_annotations: Optional[bool] = False,
        show_boundaries: Optional[bool] = False,
        annotation_fontsize: int = 14,
        tick_fontsize: int = 13,
        *args,
        **kwargs,
    ) -> Optional[plt.Figure]:
        """
        Plots a DataFrame as a heatmap with the columns along the x-axis and rows along
        the y-axis.

        Parameters
        ----------
        figsize : Tuple
            tuple specifying the size of the figure. Default is (12, 8).
        x_axis_label : str
            label for x-axis.
        y_axis_label : str
            label for y-axis.
        axis_fontsize : int
            the font size for the axis labels.
        title : str
            the figure's title.
        title_fontsize : int
            the font size for the title.
        title_xadjust : float
            sets the x position of the title text.
        title_yadjust : float
            sets the y position of the title text.
        vmin : float
            optional minimum value for heatmap scale.
        vmax : float
            optional maximum value for heatmap scale.
        show : bool
            if True, the image is displayed.
        save_to_file : str
            the path at which to save the heatmap as an image. If not specified, the
            plot will not be saved.
        dpi : int
            the resolution in dots per inch used if saving the figure.
        return_figure : bool
            if True, the function will return the figure.
        on_axis : plt.Axes
            optional `plt.Axes` object to be used instead of creating a new one.
        max_xticks : int
            the maximum number of ticks to be displayed along the x axis. Default is 50.
        cmap : mpl.colors.Colormap
            string or matplotlib Colormap object specifying the colormap of the plot.
        rotate_xticks : int
            number of degrees to rotate the tick labels on the x-axis. Default is zero.
        rotate_yticks : int
            number of degrees to rotate the tick labels on the y-axis. Default is zero.
        show_tick_lines : bool
            if True, lines are shown for ticks. Default is True.
        show_colorbar : bool
            if True, the colorbar is shown. Default is True.
        show_annotations : bool
            if True, annotations display the value of each cell. Default is False.
        show_boundaries : bool
            if True, cells are divided by a grid. Default is False.
        annotation_fontsize : int
            sets the font size of the annotations.
        tick_fontsize : int
            sets the font size of tick labels.
        """

        if on_axis:
            fig: plt.Figure = on_axis.get_figure()
            ax: plt.Axes = on_axis
        else:
            fig: plt.Figure
            ax: plt.Axes
            fig, ax = plt.subplots(figsize=figsize, layout="constrained")

        data = df.to_numpy()

        im = ax.imshow(
            data,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            aspect="auto",
            **kwargs,
        )

        xtick_labels = df.columns.to_list()
        ytick_labels = df.index.to_list()

        ax.set_xticks(np.arange(len(xtick_labels)), labels=xtick_labels)
        ax.set_yticks(np.arange(len(ytick_labels)), labels=ytick_labels)

        ax.set_xticklabels(
            xtick_labels,
            rotation=rotate_xticks,
            ha="center",
            minor=False,
        )
        ax.set_yticklabels(
            ytick_labels,
            rotation=rotate_yticks,
            ha="right",
            minor=False,
            rotation_mode="anchor",
        )

        ax.tick_params(axis="both", which="major", labelsize=tick_fontsize)

        if show_tick_lines:
            ax.tick_params(which="major", length=4, width=1, direction="out")

        plt.grid(False)

        if show_boundaries:
            ax.spines[:].set_visible(False)
            ax.set_xticks(np.arange(len(xtick_labels) + 1) - 0.5, minor=True)
            ax.set_yticks(np.arange(len(ytick_labels) + 1) - 0.5, minor=True)
            ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
            ax.tick_params(which="minor", bottom=False, left=False)
        else:
            # Limits the number of ticks shown on the x-axis.
            ax.xaxis.set_major_locator(plt.MaxNLocator(nbins=max_xticks - 1))

        ax.set_title(
            title,
            fontsize=title_fontsize,
            x=title_xadjust,
            y=title_yadjust,
        )

        if x_axis_label:
            ax.set_xlabel(x_axis_label, fontsize=axis_fontsize)

        if y_axis_label:
            ax.set_ylabel(y_axis_label, fontsize=axis_fontsize)

        if show_colorbar:
            ax.figure.colorbar(im, ax=ax)

        if show_annotations:
            data = np.around(data, decimals=1)
            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    color = im.cmap(im.norm(im.get_array()))[i, j]
                    lum = relative_luminance(color)
                    text_color = ".15" if lum > 0.408 else "w"
                    text_kwargs = dict(
                        color=text_color,
                        ha="center",
                        va="center",
                        size=annotation_fontsize,
                    )
                    if not np.isnan(data[i, j]):
                        ax.text(j, i, data[i, j], **text_kwargs)

        ax.tick_params(axis="y", which="major", pad=8)

        if save_to_file:
            plt.savefig(
                save_to_file,
                dpi=dpi,
                bbox_inches="tight",
            )

        if show:
            plt.show()
            return

        if return_figure:
            return fig

    def plot_metric(
        self,
        x_axis_column,
        y_axis_column,
        metric,
        xcats=None,
        cids=None,
        start=None,
        end=None,
        freq=None,
        agg="mean",
        figsize: Optional[Tuple[Number, Number]] = (12, 8),
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 14,
        title: Optional[str] = None,
        title_fontsize: int = 22,
        title_xadjust: Number = 0.5,
        title_yadjust: Number = 1.0,
        vmin: Optional[Number] = None,
        vmax: Optional[Number] = None,
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        on_axis: Optional[plt.Axes] = None,
        max_xticks: int = 50,
        cmap: Optional[Union[str, mpl.colors.Colormap]] = None,
        rotate_xticks: Optional[Number] = 0,
        rotate_yticks: Optional[Number] = 0,
        show_tick_lines: Optional[bool] = True,
        show_colorbar: Optional[bool] = True,
        show_annotations: Optional[bool] = False,
        show_boundaries: Optional[bool] = False,
        annotation_fontsize: int = 14,
        tick_fontsize: int = 13,
        *args,
        **kwargs,
    ):
        """
        Plots a metric from the DataFrame as a heatmap.

        Parameters
        ----------
        x_axis_column : str
            the column to be used as the x-axis.
        y_axis_column : str
            the column to be used as the y-axis.
        metric : str
            the metric to be plotted.
        xcats : List[str]
            a list of xcats to select from the DataFrame. If None, all xcats are selected.
        cids : List[str]
            a list of cids to select from the DataFrame. If None, all cids are selected.
        start : str
            ISO-8601 formatted date string. Select data from this date onwards. If None,
            all dates are selected.
        end : str
            ISO-8601 formatted date string. Select data up to and including this date. If
            None, all dates are selected.
        freq : str
            frequency to downsample the data. Default is None.
        agg : str
            aggregation method. Must be one of 'mean', 'median', 'min', 'max', 'first' or
            'last'.
        figsize : Tuple[float, float]
            tuple specifying the size of the figure. Default is (12, 8).
        x_axis_label : str
            label for x-axis.
        y_axis_label : str
            label for y-axis.
        axis_fontsize : int
            the font size for the axis labels.
        title : str
            the figure's title.
        title_fontsize : int
            the font size for the title.
        title_xadjust : float
            sets the x position of the title text.
        title_yadjust : float
            sets the y position of the title text.
        vmin : float
            optional minimum value for heatmap scale.
        vmax : float
            optional maximum value for heatmap scale.
        show : bool
            if True, the image is displayed.
        save_to_file : str
            the path at which to save the heatmap as an image. If not specified, the
            plot will not be saved.
        dpi : int
            the resolution in dots per inch used if saving the figure.
        return_figure : bool
            if True, the function will return the figure.
        on_axis : plt.Axes
            optional `plt.Axes` object to be used instead of creating a new one.
        max_xticks : int
            the maximum number of ticks to be displayed along the x axis. Default is 50.
        cmap : mpl.colors.Colormap
            string or matplotlib Colormap object specifying the colormap of the plot.
        rotate_xticks : int
            number of degrees to rotate the tick labels on the x-axis. Default is zero.
        rotate_yticks : int
            number of degrees to rotate the tick labels on the y-axis. Default is zero.
        show_tick_lines : bool
            if True, lines are shown for ticks. Default is True.
        show_colorbar : bool
            if True, the colorbar is shown. Default is True.
        show_annotations : bool
            if True, annotations display the value of each cell. Default is False.
        show_boundaries : bool
            if True, cells are divided by a grid. Default is False.
        annotation_fontsize : int
            sets the font size of the annotations.
        tick_fontsize : int
            sets the font size of tick labels.
        """

        df = self.df.copy()
        if not xcats:
            xcats = self.xcats
        if not cids:
            cids = self.cids
        if not start:
            start = self.start
        if not end:
            end = self.end

        # Validation checks not covered by Plotter.
        if metric not in ["value", "eop_lag", "mop_lag", "grading"]:
            raise ValueError(
                "`metric` must be either 'eop_lag', 'mop_lag', 'grading', or 'value'"
            )

        if not isinstance(agg, str):
            raise TypeError("`agg` must be a string")
        else:
            agg: str = agg.lower()
            if agg not in ["mean", "median", "min", "max", "first", "last"]:
                raise ValueError(
                    "`agg` must be one of 'mean', 'median', 'min', 'max', 'first' or "
                    "'last'"
                )

        df = df[["xcat", "cid", "real_date", metric]]

        if freq:
            df: pd.DataFrame = downsample_df_on_real_date(
                df=df, groupby_columns=["cid", "xcat"], freq=freq, agg=agg
            )

        if "real_date" not in [x_axis_column, y_axis_column]:
            df = df.groupby(["xcat", "cid"], observed=True).mean().reset_index()
        else:
            df["real_date"] = df["real_date"].dt.strftime("%Y-%m-%d")

        vmax: float = max(1, df[metric].max())
        vmin: float = min(0, df[metric].min())

        df = df.pivot_table(
            index=y_axis_column,
            columns=x_axis_column,
            values=metric,
            observed=True,
        )

        if figsize is None:
            figsize = (
                max(df.shape[0] / 2, 15),
                max(1, df.shape[1] / 2),
            )
        elif isinstance(figsize, list):
            figsize = tuple(figsize)

        self.plot(
            df=df,
            figsize=figsize,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            axis_fontsize=axis_fontsize,
            title=title,
            title_fontsize=title_fontsize,
            title_xadjust=title_xadjust,
            title_yadjust=title_yadjust,
            vmin=vmin,
            vmax=vmax,
            show=show,
            save_to_file=save_to_file,
            dpi=dpi,
            return_figure=return_figure,
            on_axis=on_axis,
            max_xticks=max_xticks,
            cmap=cmap,
            rotate_xticks=rotate_xticks,
            rotate_yticks=rotate_yticks,
            show_tick_lines=show_tick_lines,
            show_colorbar=show_colorbar,
            show_annotations=show_annotations,
            show_boundaries=show_boundaries,
            annotation_fontsize=annotation_fontsize,
            tick_fontsize=tick_fontsize,
        )


if __name__ == "__main__":
    test_cids: List[str] = [
        "USD",
    ]  # "EUR", "GBP"]
    test_xcats: List[str] = ["FX", "IR"]
    dfE: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="sharp-hill"
    )

    dfM: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="four-bit-sine"
    )

    dfG: pd.DataFrame = make_test_df(cids=test_cids, xcats=test_xcats, style="sine")

    dfE.rename(columns={"value": "eop_lag"}, inplace=True)
    dfM.rename(columns={"value": "mop_lag"}, inplace=True)
    dfG.rename(columns={"value": "grading"}, inplace=True)
    mergeon = ["cid", "xcat", "real_date"]
    dfx: pd.DataFrame = pd.merge(pd.merge(dfE, dfM, on=mergeon), dfG, on=mergeon)

    heatmap = Heatmap(df=dfx, xcats=["FX"])

    heatmap.df["real_date"] = heatmap.df["real_date"].dt.strftime("%Y-%m-%d")
    heatmap.df = heatmap.df.pivot_table(
        index="cid", columns="real_date", values="grading"
    )

    heatmap.plot(heatmap.df, title="abc", rotate_xticks=90)



from .qdf import QuantamentalDataFrame, QuantamentalDataFrameBase
from .generic import NoneType, SubscriptableMeta, ArgValidationMeta

__all__ = [
    "QuantamentalDataFrame",
    "QuantamentalDataFrameBase",
    "NoneType",
    "SubscriptableMeta",
    "ArgValidationMeta",
]



from .weighted_ls_regressors import (
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
)

from .modified_ls_regressors import (
    ModifiedLinearRegression,
    ModifiedSignWeightedLinearRegression,
    ModifiedTimeWeightedLinearRegression,
)

__all__ = [
    "SignWeightedLinearRegression",
    "TimeWeightedLinearRegression",
    "ModifiedLinearRegression",
    "ModifiedSignWeightedLinearRegression",
    "ModifiedTimeWeightedLinearRegression",
]


import numbers

import numpy as np
import pandas as pd
from sklearn.linear_model import Lars, lars_path

from macrosynergy.learning.preprocessing.panel_selectors.base_panel_selector import (
    BasePanelSelector,
)
from macrosynergy.learning.random_effects import RandomEffects


class LarsSelector(BasePanelSelector):
    def __init__(self, n_factors=10, fit_intercept=False):
        """
        Statistical feature selection using LARS.

        Parameters
        ----------
        n_factors : int, default=10
            Number of factors to select.
        fit_intercept : bool, default=False
            Whether to fit an intercept term in the LARS model.

        Notes
        -----
        The Least Angle Regression (LARS) algorithm was designed to fit high dimensional
        linear models. It is a means of estimating the covariates to include in the model,
        as well as associated coefficients. LARS can be considered to be a continuous
        equivalent to forward selection.

        The algorithm is described in detail in [1]_ and is implemented in the
        `scikit-learn` library [2]_. It works as follows:

        1. Set coefficients to zero.
        2. Find the covariate that has the highest correlation with the target variable.
        3. Increase the coefficient of this covariate in a stepwise fashion, recording
            the residual at each step. Stop when another covariate is as correlated with
            the residuals as the current one.
        4. Add this second covariate to the model and the compute the two-variable OLS
            solution.
        5. Increase the coefficients of the two covariates in a stepwise fashion towards
            the OLS solution, recording the residuals at each step. Stop when another
            covariate is as correlated with the residuals as the current ones.
        6. Add this third covariate to the model and compute the three-variable OLS
            solution.
        7. Iterate this process until the desired number of covariates have been selected.

        References
        ----------
        .. [1] Efron, B., Hastie, T., Johnstone, I. and Tibshirani, R., 2004.
            Least angle regression.
            https://arxiv.org/abs/math/0406456
        .. [2] https://scikit-learn.org/dev/modules/linear_model.html#least-angle-regression

        """
        # Checks
        if not isinstance(fit_intercept, bool):
            raise TypeError("'fit_intercept' must be a boolean.")
        if not isinstance(n_factors, int):
            raise TypeError("'n_factors' must be an integer.")
        if n_factors <= 0:
            raise ValueError("'n_factors' must be a positive integer.")

        # Attributes
        self.fit_intercept = fit_intercept
        self.n_factors = n_factors

        super().__init__()

    def determine_features(self, X, y):
        """
        Create feature mask based on the LARS algorithm.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame
            The target vector.

        Returns
        -------
        mask : list
            Boolean mask of selected features.
        """
        lars = Lars(fit_intercept=self.fit_intercept, n_nonzero_coefs=self.n_factors)
        lars.fit(X.values, y.values.reshape(-1, 1))
        coefs = lars.coef_

        return [True if coef != 0 else False for coef in coefs]


class LassoSelector(BasePanelSelector):
    def __init__(self, n_factors=10, positive=False):
        """
        Statistical feature selection with LASSO-LARS.

        Parameters
        ----------
        n_factors : int
            Number of factors to select.
        positive : bool
            Whether to constrain the LASSO coefficients to be positive.

        Notes
        -----
        The Least Absolute Shrinkage and Selection Operator (LASSO) [1]_ is a linear model
        that estimates sparse coefficients. This means that some encouragement is given
        for the model to set some coefficients to zero. Hence, the LASSO can be said to
        perform feature selection. It transpires that the LARS algorithm [2]_
        (see `LarsSelector`) can be used to track the LASSO coefficients as the user-defined
        sparsity level is increased. Consequently, we use the LARS algorithm to
        compute the LASSO paths and select the desired number of factors. See [3]_ for
        the `scikit-learn` documentation on the LASSO-LARS model fit.

        References
        ----------
        .. [1] Tibshirani, R., 1996. Regression shrinkage and selection via the lasso.
            Journal of the Royal Statistical Society Series B: Statistical Methodology,
            58(1), pp.267-288.
            https://www.jstor.org/stable/2346178
        .. [2] Efron, B., Hastie, T., Johnstone, I. and Tibshirani, R., 2004.
            Least angle regression.
            https://arxiv.org/abs/math/0406456
        .. [3] https://scikit-learn.org/dev/modules/generated/sklearn.linear_model.LassoLars.html
        """
        # Checks
        if not isinstance(n_factors, int):
            raise TypeError("'n_factors' must be an integer.")
        if n_factors <= 0:
            raise ValueError("'n_factors' must be a positive integer.")
        if not isinstance(positive, bool):
            raise TypeError("'positive' must be a boolean.")

        # Attributes
        self.n_factors = n_factors
        self.positive = positive

        super().__init__()

    def determine_features(self, X, y):
        """
        Create feature mask based on the LASSO-LARS algorithm.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame
            The target vector.

        Returns
        -------
        mask : np.ndarray
            Boolean mask of selected features.
        """
        # Obtain coefficient paths with dimensions (n_features, n_alphas)
        _, _, coefs_path = lars_path(
            X.values,
            y.values,
            positive=self.positive,
            method="lasso",
        )

        mask = coefs_path[:, min(self.n_factors, coefs_path.shape[1] - 1)] != 0

        return mask


class MapSelector(BasePanelSelector):
    def __init__(self, n_factors=None, significance_level=0.05, positive=False):
        """
        Univariate statistical feature selection using the Macrosynergy panel test.

        Parameters
        ----------
        n_factors : int, optional
            Number of factors to select.
        significance_level : float, default=0.05
            Significance level.
        positive : bool, default=False
            Whether to only keep features with positive estimated model coefficients.

        Notes
        -----
        The Macrosynergy panel test [1]_ is a univariate test that estimates the
        significance of a relationship between each feature and the target variable, over
        a panel. This test accounts for cross-sectional correlations. Often, different
        cross-sections in a panel are highly correlated - particularly in the case of
        dependent variable return data. This violates the assumption of independence
        in the usual z-test or t-test, from which the usual p-values are derived. As a
        consequence, probabilities of significance can be overstated.

        In the Macrosynergy panel test, a Wald test is used to compare the null hypothesis
        of an intercept + period-specific random effects model against the alternative
        hypothesis of an intercept + period-specific random effects model + the feature
        of interest. This works because the null-alternative hypotheses are nested models.
        The model in the null hypothesis accounts for the cross-sectional correlations
        that exist in each time period. Rejecting this model in favour of the alternative
        model indicates that the feature of interest is significant, accounting for those
        cross-sectional correlations.

        References
        ----------
        .. [1] Gholkar, Rushil and Sueppel, Ralph, 2023.
            Testing macro trading factors.
            https://macrosynergy.com/research/testing-macro-trading-factors/
        """
        # Checks
        if n_factors is not None:
            if not isinstance(n_factors, int):
                raise TypeError("The 'n_factors' parameter must be an integer.")
            if n_factors <= 0:
                raise ValueError(
                    "The 'n_factors' parameter must be a positive integer."
                )
        if not isinstance(significance_level, numbers.Number):
            raise TypeError("The significance_level must be a float.")
        if (significance_level < 0) or (significance_level > 1):
            raise ValueError("The significance_level must be in between 0 and 1.")
        if not isinstance(positive, (bool, np.bool_)):
            raise TypeError("The 'positive' parameter must be a boolean.")

        self.significance_level = significance_level
        self.positive = positive
        self.n_factors = n_factors

    def determine_features(self, X, y):
        """
        Create feature mask based on the Macrosynergy panel test.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame
            The target vector.

        Returns
        -------
        mask : np.ndarray
            Boolean mask of selected features.
        """
        # Iterate through each feature and perform the panel test
        factor_pvals = []

        for col in self.feature_names_in_:
            ftr = X[col]
            re = RandomEffects(fit_intercept=True).fit(ftr, y)
            est = re.params[col]
            pval = re.pvals[col]
            factor_pvals.append(pval)

        if self.n_factors is not None:
            # Return a mask of factors with `n_factors` smallest p_values
            factor_indexes = np.argsort(factor_pvals)[: self.n_factors]
            mask = [
                True if idx in factor_indexes else False
                for idx in range(len(factor_pvals))
            ]

        else:
            if self.positive:
                # Return a mask of factors with positive estimated coefficients and
                # p_values < significance_level
                mask = [
                    True if ((est > 0) and (pval < self.significance_level)) else False
                    for est in factor_pvals
                ]
            else:
                # Return as mask of factors with p_values < significance_level
                mask = [
                    True if pval < self.significance_level else False
                    for pval in factor_pvals
                ]

        return np.array(mask)


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf

    # Randomly generate a panel
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()
    train = train[
        train.index.get_level_values(1) >= pd.Timestamp(year=2005, month=8, day=1)
    ]

    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    # LARS selector
    lars = LarsSelector(n_factors=2).fit(X_train, y_train)
    print(f"LARS 2-factors, no intercept: {lars.get_feature_names_out()}")
    lars = LarsSelector(n_factors=2, fit_intercept=True).fit(X_train, y_train)
    print(f"LARS 2-factors, with intercept: {lars.get_feature_names_out()}")

    print(lars.transform(X_train))

    # LASSO selector
    lasso = LassoSelector(n_factors=1, positive=True).fit(X_train, y_train)
    print(f"Lasso 1-factor, positive restriction: {lasso.get_feature_names_out()}")
    lasso = LassoSelector(n_factors=3, positive=False).fit(X_train, y_train)
    print(f"Lasso 3-factors, with intercept: {lasso.get_feature_names_out()}")

    print(lasso.transform(X_train))

    # Map selector
    map_selector = MapSelector(n_factors=2).fit(X_train, y_train)
    print(f"Map 2-factors: {map_selector.get_feature_names_out()}")
    map_selector = MapSelector(significance_level=0.2).fit(X_train, y_train)
    print(f"Map significance 0.2: {map_selector.get_feature_names_out()}")

    print(map_selector.transform(X_train))



"""
Implementation of panel calculation functions for quantamental data. The functionality
allows applying mathematical operations on time-series data.
"""

import numpy as np
import pandas as pd
from typing import List, Tuple
from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.utils import drop_nan_series
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy import PYTHON_3_8_OR_LATER
import re
import random


def panel_calculator(
    df: pd.DataFrame,
    calcs: List[str] = None,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    external_func: dict = None,
) -> pd.DataFrame:
    """
    Calculates new data panels through a given input formula which is performed on
    existing panels.

    Parameters
    ----------
    df : ~pandas.Dataframe
        standardized dataframe with following necessary columns: 'cid', 'xcat',
        'real_date' and 'value'.
    calcs : List[str]
        list of formulas denoting operations on panels of categories. Words in capital
        letters denote category panels. Otherwise the formulas can include numpy functions
        and standard binary operators. See notes below.
    cids : List[str]
        cross sections over which the panels are defined.
    start : str
        earliest date in ISO format. Default is None and earliest date in df is used.
    end : str
        latest date in ISO format. Default is None and latest date in df is used.
    blacklist : dict
        cross sections with date ranges that should be excluded from the dataframe. If
        one cross section has several blacklist periods append numbers to the cross-section
        code.
    external_func : dict
        dictionary of external functions to be used in the panel calculation. The key is
        the name of the function and the value is the function object itself. e.g.
        {"my_func": my_func}.

    Returns
    -------
    ~pandas.Dataframe
        standardized dataframe with all new categories in standard format, i.e the
        columns 'cid', 'xcat', 'real_date' and 'value'.


    Notes
    -----
    Panel calculation strings can use numpy functions and unary/binary operators on
    category panels. The category is indicated by capital letters, underscores and
    numbers. Panel category names that are not at the beginning or end of the string
    must always have a space before and after the name. Calculated category and
    panel operations must be separated by '='.

    Examples:

    .. code-block:: python

        NEWCAT = ( OLDCAT1 + 0.5) * OLDCAT2
    or

    .. code-block:: python

        NEWCAT = np.log( OLDCAT1 ) - np.abs( OLDCAT2 ) ** 1/2

    Panel calculation can also involve individual indicator
    series (to be applied to all series in the panel by using th 'i' as prefix), such
    as:

    .. code-block:: python

        NEWCAT = OLDCAT1 - np.sqrt( iUSD_OLDCAT2 )

    These strings are passed as a list of strings (`calcs`) to the function.

    If more than one new category is calculated, the resulting panels can be used
    sequentially in the calculations, such as:
    .. code-block:: python

        ["NEWCAT1 = 1 + OLDCAT1 / 100", "NEWCAT2 = OLDCAT2 * NEWCAT1"]

    .. code-block:: python

        calcs = [
            "NEWCAT = OLDCAT1 + OLDCAT2",
            "NEWCAT2 = CAT_A * CAT_B - CAT_C * 0.5",
            "NEWCAT3 = OLDCAT1 - np.sqrt(iUSD_OLDCAT2)",
        ]

        df = panel_calculator(df=df, calcs=calcs, ...)
    """

    # A. Asserts

    cols = ["cid", "xcat", "real_date", "value"]

    col_error = f"The DataFrame must contain the necessary columns: {cols}."
    assert set(cols).issubset(set(df.columns)), col_error
    # Removes any columns beyond the required.
    df = QuantamentalDataFrame(df[cols])
    _as_categorical = df.InitializedAsCategorical
    assert isinstance(calcs, list), "List of functions expected."

    error_formula = "Each formula in the panel calculation list must be a string."
    assert all([isinstance(elem, str) for elem in calcs]), error_formula
    assert isinstance(cids, list), "List of cross-sections expected."

    _check_calcs(calcs)

    if external_func:
        globals().update(external_func)

    # B. Collect new category names and their formulas.

    ops = {}
    for calc in calcs:
        calc_parts = calc.split("=", maxsplit=1)
        ops[calc_parts[0].strip()] = calc_parts[1].strip()

    # C. Check if all required categories are in the dataframe.

    old_xcats_used, singles_used, single_cids = _get_xcats_used(ops)

    old_xcats_used = list(set(old_xcats_used))
    missing = sorted(set(old_xcats_used) - set(df["xcat"].unique()))

    new_xcats = list(ops.keys())
    if len(missing) > 0 and not set(missing).issubset(set(new_xcats)):
        raise ValueError(f"Missing categories: {missing}.")

    # If any of the elements of single_cids are not in cids, add them to cids.
    cids_used = list(set(single_cids + cids))

    # D. Reduce dataframe with intersection requirement.

    dfx = reduce_df(
        df,
        xcats=old_xcats_used,
        cids=cids_used,
        start=start,
        end=end,
        blacklist=blacklist,
        intersect=False,
    )

    # E. Create all required wide dataframes with category names.
    df = df.add_ticker_column()
    for xcat in old_xcats_used:
        dfxx = dfx[dfx["xcat"] == xcat]
        dfw = dfxx.pivot(index="real_date", columns="cid", values="value")
        dfw = _replace_zeros(df=dfw)
        exec(f"{xcat} = dfw")

    for single in singles_used:
        ticker = single[1:]
        dfxx = df[(df["ticker"]) == ticker]
        if dfxx.empty:
            raise ValueError(f"Ticker, {ticker}, missing from the dataframe.")
        else:
            dfx1 = dfxx.set_index("real_date")["value"].to_frame()
            dfx1 = dfx1.truncate(before=start, after=end)

            dfw = pd.concat([dfx1] * len(cids), axis=1, ignore_index=True)
            dfw.columns = cids
            dfw = _replace_zeros(df=dfw)
            exec(f"{single} = dfw")

    # F. Calculate the panels and collect.
    df_out: pd.DataFrame
    for new_xcat, formula in ops.items():
        dfw_add = eval(formula)
        df_add = pd.melt(dfw_add.reset_index(), id_vars=["real_date"]).rename(
            {"variable": "cid"}, axis=1
        )
        df_add = QuantamentalDataFrame.from_long_df(df_add, xcat=new_xcat)
        if new_xcat == list(ops.keys())[0]:
            df_out = df_add[cols]
        else:
            df_out = pd.concat([df_out, df_add[cols]], axis=0, ignore_index=True)
        dfw_add = _replace_zeros(df=dfw_add)
        exec(f"{new_xcat} = dfw_add")

    if df_out.isna().any().any():
        df_out = drop_nan_series(df=df_out, raise_warning=True)

    df_out = QuantamentalDataFrame(df_out, categorical=_as_categorical)
    return df_out


def time_series_check(formula: str, index: int):
    """
    Determine if the panel has any time-series methods applied. If a time-series
    conversion is applied, the function will return the terminal index of the respective
    category. Further, a boolean parameter is also returned to confirm the presence of a
    time-series operation.

    Parameters
    ----------
    formula : str

    index : int
        starting index to iterate over.

    Returns
    -------
    Tuple[int, bool]
    """

    check = lambda a, b, c: (
        (a.isupper() or a.isnumeric()) and b == "." and c.islower()
    )

    f = formula
    length = len(f)
    clause = False
    for i in range(index, (length - 2)):
        if check(f[i], f[i + 1], f[i + 2]):
            clause = True
            break

    return i, clause


def xcat_isolator(expression: str, start_index: str, index: int) -> Tuple[str, int]:
    """
    Split the category from the time-series operation. The function will return the
    respective category.

    Parameters
    ----------
    expression : str

    start_index : str
        starting index to search over.
    index : int
        defines the end of the search space over the expression.

    Returns
    -------
    Tuple[str, int]
        xcat string, and the string index where the xcat ends.
    """

    op_copy = expression[start_index : index + 1]

    start = next(i for i, elem in enumerate(op_copy) if elem.isupper())

    xcat = op_copy[start : index + 1]

    return xcat, start_index + start + len(xcat)


def _get_xcats_used(ops: dict) -> Tuple[List[str], List[str]]:
    """
    Collect all categories used in the panel calculation.

    Parameters
    ----------
    ops : dict
        dictionary of panel calculation formulas.

    Returns
    -------
    Tuple[List[str], List[str]]
        all_xcats_used, singles_used.
    """

    xcats_used: List[str] = []
    singles_used: List[str] = []
    for op in ops.values():
        index, clause = time_series_check(formula=op, index=0)
        start_index = 0
        if clause:
            while clause:
                xcat, end_ = xcat_isolator(op, start_index, index)
                xcats_used.append(xcat)
                index, clause = time_series_check(op, index=end_)
                start_index = end_
        else:
            op_list = op.split(" ")
            xcats_used += [x for x in op_list if re.match("^[A-Z]", x)]
            singles_used += [s for s in op_list if re.match("^i", s)]

    single_xcats = [x.split("_", 1)[1] for x in singles_used]
    single_cids = [x.split("_", 1)[0] for x in single_xcats]
    all_xcats_used = xcats_used + single_xcats
    return all_xcats_used, singles_used, single_cids


def _check_calcs(formulas: List[str]):
    """
    Check formulas for invalid characters in xcats.

    Parameters
    ----------
    calcs : List[str]
        list of formulas.

    Returns
    -------
    List[str]
        list of formulas.
    """

    pattern = r"[-+*()/](?=i?[A-Z])|(?<=[A-Z])[-+*()/]"

    for formula in formulas:
        for term in formula.split():
            # Search for any occurrences of the pattern in the input string
            if re.search(pattern, term):
                raise ValueError(
                    f"Invalid character found next to a capital letter or 'i' in string: {term}. "
                    + "Arithmetic operators and parentheses must be separated by spaces."
                )


def _replace_zeros(df: pd.DataFrame):
    """
    Replace zeros with NaNs in the dataframe.

    Parameters
    ----------
    df : ~pandas.DataFrame
        dataframe to be cleaned.

    Returns
    -------
    ~pandas.DataFrame
        cleaned dataframe.
    """

    if not PYTHON_3_8_OR_LATER: # pragma: no cover
        for col in df.columns:
            df[col] = df[col].replace(pd.NA, np.nan)
            df[col] = df[col].astype("float64")
        return df
    else:
        return df

    return df


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["EUR"] = ["2002-01-01", "2020-09-30", -0.2, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2010-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2012-01-01", "2020-09-30", 1, 2, 0.8, 0.5]

    random.seed(2)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Example blacklist.
    black = {"AUD": ["2000-01-01", "2003-12-31"]}

    start = "2010-01-01"
    end = "2020-12-31"

    # Example filter for dataframe.
    filt1 = (dfd["xcat"] == "XR") | (dfd["xcat"] == "CRY")
    dfdx = dfd[filt1]

    # First testcase.

    f1 = "NEW_VAR1 = GROWTH - iEUR_INFL"
    formulas = [f1]
    cidx = ["AUD", "CAD"]
    df_calc = panel_calculator(
        df=dfd, calcs=formulas, cids=cidx, start=start, end=end, blacklist=black
    )
    # Second testcase: EUR is not passed in as one of the cross-sections in "cids"
    # parameter but is defined in the dataframe. Therefore, code will not break.
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    formula = "NEW1 = XR - iUSD_XR"
    formula_2 = "NEW2 = GROWTH - iEUR_INFL"
    formulas = [formula, formula_2]
    df_calc = panel_calculator(
        df=dfd, calcs=formulas, cids=cids, start=start, end=end, blacklist=black
    )



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from typing import List, Optional
from numbers import Number

from macrosynergy.download.transaction_costs import (
    download_transaction_costs,
    AVAIALBLE_COSTS,
    AVAILABLE_STATS,
)
from macrosynergy.management.utils import (
    reduce_df,
    get_cid,
    get_xcat,
)
from macrosynergy.management.types import QuantamentalDataFrame


def get_fids(df: QuantamentalDataFrame) -> list:
    def repl(x: str, yL: List[str]) -> str:
        for y in yL:
            x = x.replace(y, "")
        return x

    fid_endings = [f"{t}_{s}" for t in AVAIALBLE_COSTS for s in AVAILABLE_STATS]
    tickers = QuantamentalDataFrame(df).list_tickers()

    return list(set(map(lambda x: repl(x, fid_endings), tickers)))


def check_df_for_txn_stats(
    df: QuantamentalDataFrame,
    fids: List[str],
    tcost_n: str,
    rcost_n: str,
    size_n: str,
    tcost_l: str,
    rcost_l: str,
    size_l: str,
) -> None:
    expected_tickers = [
        f"{_fid}{txn_ticker}"
        for _fid in fids
        for txn_ticker in [tcost_n, rcost_n, size_n, tcost_l, rcost_l, size_l]
    ]
    found_tickers = QuantamentalDataFrame(df).list_tickers()
    if not set(expected_tickers).issubset(set(found_tickers)):
        raise ValueError(
            "The dataframe is missing the following tickers: "
            + ", ".join(set(expected_tickers) - set(found_tickers))
        )


def get_diff_index(df_wide: pd.DataFrame, freq: str = "D") -> pd.Index:
    df_diff = df_wide.diff(axis=0)
    change_index = df_diff.index[((df_diff.abs() > 0) | df_diff.isnull()).any(axis=1)]
    return change_index


def extrapolate_cost(
    trade_size: Number,
    median_size: Number,
    median_cost: Number,
    pct90_size: Number,
    pct90_cost: Number,
) -> Number:
    err_msg = "`{k}` must be a number > 0"
    if not isinstance(trade_size, Number):
        raise TypeError(err_msg.format(k="trade_size"))
    trade_size = abs(trade_size)

    for k, v in [
        ("trade_size", trade_size),
        ("median_size", median_size),
        ("median_cost", median_cost),
        ("pct90_size", pct90_size),
        ("pct90_cost", pct90_cost),
    ]:
        if not isinstance(v, Number):
            raise TypeError(err_msg.format(k=k))
        if v < 0:
            raise ValueError(err_msg.format(k=k))

    if trade_size <= median_size:
        cost = median_cost
    else:
        b = (pct90_cost - median_cost) / (pct90_size - median_size)
        cost = median_cost + b * (trade_size - median_size)
    return cost


def _plot_costs_func(
    tco: "TransactionCosts",
    fids: Optional[List[str]],
    cost_type: str,
    ncol: int,
    x_axis_label: str,
    y_axis_label: str,
    title: Optional[str] = None,
    title_fontsize: int = 28,
    facet_title_fontsize: int = 20,
    *args,
    **kwargs,
):
    tco.check_init()
    if fids is None:
        fids = tco.fids
    if not isinstance(fids, list) or not all(isinstance(fid, str) for fid in fids):
        raise ValueError("`fids` must be a list of strings")

    costfunc = tco.bidoffer if cost_type == "BIDOFFER" else tco.rollcost

    nrows = len(fids) // ncol + (len(fids) % ncol > 0)
    sns.set_theme(style="whitegrid")
    fig, axes = plt.subplots(
        nrows=nrows, ncols=ncol, figsize=(5 * ncol, 5 * nrows), layout="constrained"
    )
    if title is None:
        fig.suptitle(f"{cost_type.capitalize()}", fontsize=28)
    else:
        fig.suptitle(title, fontsize=title_fontsize)

    # Define colors for each date range
    colors = sns.color_palette("viridis", n_colors=len(tco.change_index))

    idx_dates = tco.change_index.tolist()

    def label_fmt(x):
        return pd.Timestamp(x).strftime("%Y-%m-%d")

    labels = [
        f"{label_fmt(d1)} to {label_fmt(d2 - pd.offsets.BDay(1))}"
        for d1, d2 in zip(idx_dates[:-1], idx_dates[1:])
    ]
    labels.append(f"{label_fmt(idx_dates[-1])} to Present")
    color_map = dict(zip(labels, colors))

    legend_handles = {}
    ax: plt.Axes
    for i, fid in enumerate(sorted(fids)):
        r, c = divmod(i, ncol)
        ax = axes[r, c] if nrows > 1 else (axes[c] if ncol > 1 else axes)
        max_trade_size = tco.df_wide[fid + tco.size_l].max()
        trade_sizes = np.arange(1, max_trade_size + 101, 1)

        for dt, lb in zip(idx_dates, labels):
            trade_costs = [
                costfunc(fid=fid, trade_size=ts, real_date=dt) for ts in trade_sizes
            ]
            line = sns.lineplot(
                x=trade_sizes,
                y=trade_costs,
                ax=ax,
                color=color_map[lb],
                label=lb,
                zorder=10,
            )

            median_trade_size = tco.df_wide.loc[dt, fid + tco.size_n]
            large_trade_size = tco.df_wide.loc[dt, fid + tco.size_l]
            median_xcost = tco.df_wide.loc[
                dt, fid + (tco.tcost_n if cost_type == "BIDOFFER" else tco.rcost_n)
            ]
            large_xcost = tco.df_wide.loc[
                dt, fid + (tco.tcost_l if cost_type == "BIDOFFER" else tco.rcost_l)
            ]
            sns.scatterplot(
                x=[median_trade_size, large_trade_size],
                y=[median_xcost, large_xcost],
                ax=ax,
                color="red",
                zorder=20,
            )

            ax.set_xlim(left=0)
            ax.set_title(f"{fid}", fontsize=facet_title_fontsize)

            if lb not in legend_handles:
                legend_handles[lb] = line.lines[0]

    # Remove individual subplot legends
    if isinstance(axes, plt.Axes):
        axes.get_legend().remove()
    else:
        for ax in axes.flat[1:]:
            if ax.get_legend() is not None:
                ax.get_legend().remove()

    fig.supxlabel(x_axis_label)
    fig.supylabel(y_axis_label)
    plt.show()


class SparseCosts(object):
    def __init__(self, df):
        if not isinstance(df, QuantamentalDataFrame):
            raise TypeError("df must be a QuantamentalDataFrame")
        self.df: QuantamentalDataFrame = QuantamentalDataFrame(df)
        self.prepare_data()

    def prepare_data(self):
        """
        Prepares data for use within the class, including setting up the wide DataFrame
        and fids. This method can be called again to refresh the data and cache.
        """

        df_wide = QuantamentalDataFrame(self.df).to_wide()
        self._all_fids = get_fids(self.df)
        change_index = get_diff_index(df_wide)  # drop rows with no change
        self.change_index: pd.DatetimeIndex = change_index
        df_wide = df_wide.loc[change_index]
        self.df_wide = df_wide

    def get_costs(self, fid: str, real_date: str) -> pd.Series:
        """
        Returns the costs for a given FID and date.

        Parameters
        ----------
        fid : str
            The FID (financial contract identifier) to get costs for.
        real_date : str
            The date to get costs for.
        """
        if fid not in self._all_fids:
            return None
        cost_names = [col for col in self.df_wide.columns if col.startswith(fid)]
        if not cost_names:
            return None
        df_loc = self.df_wide.loc[:real_date, cost_names]
        last_valid_index = df_loc.last_valid_index()
        return df_loc.loc[last_valid_index] if last_valid_index is not None else None


class TransactionCosts(object):
    """
    Interface to query transaction statistics dataframe.
    """

    DEFAULT_ARGS = dict(
        tcost_n="BIDOFFER_MEDIAN",
        rcost_n="ROLLCOST_MEDIAN",
        size_n="SIZE_MEDIAN",
        tcost_l="BIDOFFER_90PCTL",
        rcost_l="ROLLCOST_90PCTL",
        size_l="SIZE_90PCTL",
    )

    def check_init(self) -> bool:
        if not hasattr(self, "sparse_costs") or not hasattr(
            self.sparse_costs, "df_wide"
        ):
            raise ValueError("The TransactionCosts object has not been initialised")
        return True

    def __init__(
        self,
        df: QuantamentalDataFrame,
        fids: List[str],
        tcost_n: str = "BIDOFFER_MEDIAN",
        rcost_n: str = "ROLLCOST_MEDIAN",
        size_n: str = "SIZE_MEDIAN",
        tcost_l: str = "BIDOFFER_90PCTL",
        rcost_l: str = "ROLLCOST_90PCTL",
        size_l: str = "SIZE_90PCTL",
    ) -> None:
        df = QuantamentalDataFrame(df)
        check_df_for_txn_stats(
            df=df,
            fids=fids,
            tcost_n=tcost_n,
            rcost_n=rcost_n,
            size_n=size_n,
            tcost_l=tcost_l,
            rcost_l=rcost_l,
            size_l=size_l,
        )
        self.fids = sorted(set(fids))
        self.tcost_n = tcost_n
        self.rcost_n = rcost_n
        self.size_n = size_n
        self.tcost_l = tcost_l
        self.rcost_l = rcost_l
        self.size_l = size_l
        self._txn_stats = [tcost_n, rcost_n, size_n, tcost_l, rcost_l, size_l]

        _cids = list(set(get_cid(fids)))
        _xcats = [f"{xc}{tc}" for xc in set(get_xcat(fids)) for tc in self._txn_stats]

        df = reduce_df(df=df, cids=_cids, xcats=_xcats)
        # drop all nan rows
        df = df.dropna(axis=0, how="any")
        self.sparse_costs = SparseCosts(df)

    @property
    def change_index(self) -> pd.DatetimeIndex:
        self.check_init()
        return self.sparse_costs.change_index

    @property
    def df_wide(self) -> pd.DataFrame:
        self.check_init()
        return self.sparse_costs.df_wide

    @property
    def qdf(self) -> QuantamentalDataFrame:
        self.check_init()
        return self.sparse_costs.df

    @staticmethod
    def from_qdf(qdf: QuantamentalDataFrame, fids: List[str]) -> "TransactionCosts":
        return TransactionCosts(df=qdf, fids=fids, **TransactionCosts.DEFAULT_ARGS)

    @classmethod
    def download(cls) -> "TransactionCosts":
        df = download_transaction_costs(categorical=True)
        return cls(df=df, fids=get_fids(df), **cls.DEFAULT_ARGS)

    def get_costs(self, fid: str, real_date: str) -> pd.Series:
        self.check_init()
        if fid not in self.fids:
            return None
        return self.sparse_costs.get_costs(fid=fid, real_date=real_date)

    @staticmethod
    def extrapolate_cost(
        trade_size: Number,
        median_size: Number,
        median_cost: Number,
        pct90_size: Number,
        pct90_cost: Number,
    ) -> Number:
        if (not isinstance(trade_size, Number)) or np.isnan(trade_size):
            return 0.0
        return extrapolate_cost(
            trade_size=trade_size,
            median_size=median_size,
            median_cost=median_cost,
            pct90_size=pct90_size,
            pct90_cost=pct90_cost,
        )

    def bidoffer(self, fid: str, trade_size: Number, real_date: str) -> Number:
        self.check_init()
        row = self.sparse_costs.get_costs(fid=fid, real_date=real_date)
        if row is None:
            return np.nan
        d = dict(
            trade_size=trade_size,
            median_size=row[fid + self.size_n],
            median_cost=row[fid + self.tcost_n],
            pct90_size=row[fid + self.size_l],
            pct90_cost=row[fid + self.tcost_l],
        )

        return self.extrapolate_cost(**d)

    def rollcost(self, fid: str, trade_size: Number, real_date: str) -> Number:
        self.check_init()
        row = self.sparse_costs.get_costs(fid=fid, real_date=real_date)
        if row is None:
            return np.nan
        d = dict(
            trade_size=trade_size,
            median_size=row[fid + self.size_n],
            median_cost=row[fid + self.rcost_n],
            pct90_size=row[fid + self.size_l],
            pct90_cost=row[fid + self.rcost_l],
        )
        return self.extrapolate_cost(**d)

    def plot_costs(
        self,
        fids: Optional[List[str]] = None,
        cost_type: str = "BIDOFFER",
        ncol: int = 8,
        x_axis_label: str = "Trade size (USD, millions)",
        y_axis_label: str = "Percent of outright forward",
        title: Optional[str] = None,
        title_fontsize: int = 28,
        facet_title_fontsize: int = 20,
        *args,
        **kwargs,
    ):
        _plot_costs_func(
            tco=self,
            fids=fids,
            cost_type=cost_type,
            ncol=ncol,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            facet_title_fontsize=facet_title_fontsize,
            title_fontsize=title_fontsize,
            *args,
            **kwargs,
        )


class ExampleAdapter(TransactionCosts):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def extrapolate_cost(
        trade_size: Number,
        median_size: Number,
        median_cost: Number,
        pct90_size: Number,
        pct90_cost: Number,
    ) -> Number:
        # just as an example
        u = median_cost / median_size
        v = pct90_cost / pct90_size
        avg_costs = (u + v) / 2
        return trade_size * avg_costs

    def bidoffer(self, fid: str, trade_size: Number, real_date: str) -> Number:
        return super().bidoffer(fid, trade_size, real_date)

    def somecalc(
        self, fid: str, trade_size: Number, real_date: str, factor=1
    ) -> Number:
        # some random computation
        row = self.sparse_costs.get_costs(fid=fid, real_date=real_date)
        d = dict(
            trade_size=trade_size,
            median_size=row[fid + self.size_n],
            median_cost=row[fid + self.rcost_n],
            pct90_size=row[fid + self.size_l],
            pct90_cost=row[fid + self.rcost_l],
        )
        d["roll_cost"] = d["roll_cost"] * factor
        return self.extrapolate_cost(**d)


if __name__ == "__main__":
    txn_costs_obj: TransactionCosts = TransactionCosts.download()

    test_dict = {
        "GBP_FXROLLCOST_MEDIAN": 0.0022470715369672,
        "GBP_FXSIZE_MEDIAN": 50.0,
        "GBP_FXSIZE_90PCTL": 200.0,
    }

    found_costs = txn_costs_obj.get_costs(
        fid="GBP_FX", real_date="2011-01-01"
    ).to_dict()

    for k, v in test_dict.items():
        assert np.isclose(found_costs[k], v)

    txn_costs_obj.plot_costs(cost_type="ROLLCOST", fids=txn_costs_obj.fids[:16], ncol=4)
